#!/usr/bin/env python


__author__ = "Ricardo Fernandes & Bartlomiej Urbaniec"
__email__ = "icecontrols.support@cern.ch"
__copyright__ = "(C) 2012 European Organization for Nuclear Research, Geneva, Switzerland"
__version__ = "1.8"
__date__ = "$Date: 2012-03-08 14:14:57 +0100 (Thu, 08 Mar 2012) $"
__description__ = "A tool that allows the extraction and transformation of data in a flexible way"
__status__ = "Production"



# =============================================
#  IMPORT MODULES
# =============================================
try:
    import sys
except:
    print "Error when importing module 'sys'."
    sys.exit(-1)

try:
    import os
except:
    print "Error when importing module 'os'."
    sys.exit(-1)

try:
    import re
except:
    print "Error when importing module 're'."
    sys.exit(-1)
    
try:
    import subprocess
except:
    print "Error when importing module 'subprocess'."
    sys.exit(-1)   

try:
    import types
except:
    print "Error when importing module 'types'."
    sys.exit(-1)

try:
    import StringIO
except:
    print "Error when importing module 'StringIO'."
    sys.exit(-1)
    
try:
    from datetime import datetime
except:
    print "Error when importing module 'datetime'."
    sys.exit(-1)

try:
    import traceback
except:
    print "Error when importing module 'traceback'."
    sys.exit(-1)

try:
    from operator import itemgetter
except:
    print "Error when importing module 'itemgetter'."
    sys.exit(-1)     

try:
    from xml.dom import minidom
    from xml.dom import Node   
    from xml.parsers.expat import ExpatError     
except:
    print "Error when importing module 'minidom'."
    sys.exit(-1)    




# =============================================
#  GLOBAL VARIABLES
# =============================================
_LOG_MESSAGE = ""
_LOG_HANDLER = None
_LOG_FLAG = False
_QUIET_FLAG = False
_UNIQUE_FLAG = False
_SHOWAUTOSELECT_FLAG = False
_STRING = 0
_INTEGER = 1
_FLOAT = 2
_DEBUG_MODE = False



# =============================================
#  FUNCTIONS
# =============================================
def getConfig(pFile):
    """
    Description:    
       This function reads and parses a certain XML config file specified in the parameter "pFile". It returns four elements: code,
       header (or line number error), input (or column number error) and output.
    
    Parameter(s):
       pFile - string containing the name of the XML config file to be read.
    
    Return(s):    
       int - 0 (the file was successfully read and parsed), -1 (error found when reading the file) or -2 (error found when parsing
             the file - i.e. an illegal structure was found).
       list (or int) - contains the header structure defined in the file. Note that if a reading error was found (-1) then this list
                       will be empty; if a parsing error was found (-2) then it will contain the line number that caused the error.                 
       list (or int) - contains the input structure defined in the file. Note that if a reading error was found (-1) then this list
                       will be empty; if a parsing error was found (-2) then it will contain the column number that caused the error. 
       list          - contains the output structure defined in the file. Note that if a reading error was found (-1 or -2) then
                       this list will be empty. 
    
    Example:
       (...)
       code, headerList, inputList, outputList = getConfig("myConfig.xml")   # read config from XML file "myConfig.xml"
       
       if code == 0:       
          for header in headerList:
             print header   # print each element of the header structure
       
          for input in inputList:
             print input   # print each element of the input structure
             
          for output in outputList:
             print output   # print each element of the output structure             
       else:
          if code == -1:
             print "Error found when reading config file."
          else:
             print "Error found when parsing config file in line '%s' at column '%s'." % (headerList, inputList)
       (...)        
    """    

    try:
        fileHandler = open(pFile, "r")
        configText = fileHandler.read()
        fileHandler.close()
        return getConfigFromText(configText)
    except:
        return -1, [], [], []
        
    



def getConfigFromText(pText):
    """
    Description:
       This function parses a certain XML config string specified in the parameter "pText". It returns four elements: code,
       header (or line number error), input (or column number error) and output.
    
    Parameter(s):
       pText - string containing the XML config.
    
    Return(s):    
       int - 0 (the text was successfully read and parsed), -1 (error found when reading the string) or -2 (error found when parsing
             the string - i.e. an illegal structure was found).
       list (or int) - contains the header structure defined in the string. Note that if an unexpected error was found (-1) then this list
                       will be empty; if a parsing error was found (-2) then it will contain the line number that caused the error.                 
       list (or int) - contains the input structure defined in the string. Note that if an unexpected error was found (-1) then this list
                       will be empty; if a parsing error was found (-2) then it will contain the column number that caused the error. 
       list          - contains the output structure defined in the string. Note that if an error was found (-1 or -2) then
                       this list will be empty. 
    
    Example:
       (...)
       
       myConfig = '<flexextractor version = "1.0" description = "Config file to generate the recipes">\
                      <input type = "EXCEL_BINARY">\
                         <source name = "Digital Parameter" range = "10:">\
                            <data name = "Equipment" select = "1"/>\
                            <data name = "Location" select = "2"/>\
                            <data name = "Name" select = "3"/>\
                            <data name = "Latch" select = "4"/>\
                            <data name = "Value" select = "8"/>\
                            <data name = "Recipes" select = "25"/>\                                    
                         </source>\
                         <source name = "Word Parameter" range = "10:">\
                            <data name = "Equipment" select = "1"/>\
                            <data name = "Location" select = "2"/>\
                            <data name = "Name" select = "3"/>\
                            <data name = "Latch" select = "4"/>\
                            <data name = "Value" select = "13"/>\
                            <data name = "Recipes" select = "24"/>\                        
                         </source>\
                      </input>\
                      <output>\
                         <section head = "{Equipment}_{Location}_{Recipes}Recipe;{Equipment}_{Location}_{Recipes}RecipeInitial;{Equipment}_{Location}_{Recipes}RecipeInitial;{Equipment}_{Location}_{Latch};" condition = "{Recipes} != ''" join = ";" generate = "{Equipment}_{Location}_{Name}|{Value}"/>\       
                      </output>\
                  </flexextractor>\'
       
       code, headerList, inputList, outputList = getConfigFromText(myConfig)   # read config from string "myConfig"
       
       if code == 0:       
          for header in headerList:
             print header   # print each element of the header structure
       
          for input in inputList:
             print input   # print each element of the input structure
             
          for output in outputList:
             print output   # print each element of the output structure             
       else:
          if code == -1:
             print "Error found when reading config."
          else:
             print "Error found when parsing config in line '%s' at column '%s'." % (headerList, inputList)
       (...)        
    """
    
    try:   
        header = {"version": None, "description": None, "preprocess": None, "postprocess": None}
        input = {"file": None, "type": None, "parameter": None, "autoselect": None, "combine": None, "condition": None, "range": None, "source": []}
        output = {"file": None, "type": None, "parameter": None, "append": None, "condition": None, "head": None, "headfile": None, "tail": None, "tailfile": None, "split": None, "match": None, "nomatch": None, "replace": None, "multiplereplace": None, "substring": None, "unique": None, "range": None, "ordergenerate": None, "joingenerate": None, "section": []}
        
        parser = minidom.parseString(pText)
        rootNode = parser.documentElement
        
        attributeList = rootNode.attributes                    
        for attribute in attributeList.keys():
            if attribute in header:
                attributeNode = attributeList.get(attribute)
                attributeValue = attributeNode.nodeValue      
                header[attribute] = _removeComment(attributeValue.encode("utf-8"))
        
        for node in rootNode.childNodes:
            
            if node.nodeType == Node.ELEMENT_NODE:
                                
                if node.nodeName == "input":
                    attributeList = node.attributes                    
                    for attribute in attributeList.keys():
                        if attribute in input:
                            attributeNode = attributeList.get(attribute)
                            attributeValue = attributeNode.nodeValue                            
                            input[attribute] = _removeComment(attributeValue.encode("utf-8"))
                    for inputChild in node.childNodes:
                        if inputChild.nodeType == Node.ELEMENT_NODE:
                            if inputChild.nodeName == "source":
                                source = {"name": None, "alias": None, "file": None, "type": None, "parameter": None, "autoselect": None, "condition": None, "unique": None, "range": None, "order": None, "data": []}
                                attributeList = inputChild.attributes                    
                                for attribute in attributeList.keys():
                                    if attribute in source:
                                        attributeNode = attributeList.get(attribute)
                                        attributeValue = attributeNode.nodeValue                                                                      
                                        source[attribute] = _removeComment(attributeValue.encode("utf-8"))
                                for sourceChild in inputChild.childNodes:
                                    if sourceChild.nodeType == Node.ELEMENT_NODE:
                                        if sourceChild.nodeName == "data":
                                            data = {"name": None, "select": None, "regexpselect": None, "default": None, "split": None, "match": None, "nomatch": None, "replace": None, "multiplereplace": None, "substring": None, "type": None, "parameter": None}
                                            attributeList = sourceChild.attributes                    
                                            for attribute in attributeList.keys():
                                                if attribute in data:
                                                    attributeNode = attributeList.get(attribute)
                                                    attributeValue = attributeNode.nodeValue                                                    
                                                    data[attribute] = _removeComment(attributeValue.encode("utf-8"))
                                            source["data"].append(data)
                                input["source"].append(source)
                                                                    
                elif node.nodeName == "output":
                    attributeList = node.attributes                    
                    for attribute in attributeList.keys():
                        if attribute in output:
                            attributeNode = attributeList.get(attribute)
                            attributeValue = attributeNode.nodeValue                            
                            output[attribute] = _removeComment(attributeValue.encode("utf-8"))
                    output["section"].extend(_getConfigRecursive(node))                    


        # =============================================
        #  MAKE SOME REPLACEMENTS FOR FASTER EXECUTION
        # =============================================
        autoSelect = input["autoselect"]
        inputRange = input["range"]
        for source in input["source"]:
            if source["autoselect"] is None:
                source["autoselect"] = autoSelect
#            if source["range"] is None:
#                source["range"] = inputRange
            dataList = source["data"]                        
            for data in dataList:                
                type = data["type"]
                if type is None:
                    type = _STRING
                else:
                    type = type.upper()
                    if type == "INTEGER":
                        type = _INTEGER
                    else:
                        if type == "FLOAT":
                            type = _FLOAT
                        else:
                            type = _STRING                    
                data["type"] = type
        
        # =============================================
        #  CHECK IF SOURCE EXIST IN CONFIG FILE
        # =============================================        
        if len(input["source"]) == 0:
            raise NameError("Config file doesn't contain source section")
        return 0, header, input, output
    
    except NameError:
        return -3, [], [], []   
    except ExpatError, expatError:
        return -2, expatError.lineno, expatError.offset + 1, []    
    except:
        return -1, [], [], []


    

def _getConfigRecursive(pParent):
    
    result = []    
    for node in pParent.childNodes:
        if node.nodeType == Node.ELEMENT_NODE:
            if node.nodeName == "section":             
                section = {"condition": None, "head": None, "headfile": None, "tail": None, "tailfile": None, "generate": None, "generatefile": None, "default": None, "defaultfile": None, "defaultunique": None, "split": None, "match": None, "nomatch": None, "replace": None, "multiplereplace": None, "substring": None, "unique": None, "range": None, "orderhead": None, "joinhead": None, "ordergenerate": None, "joingenerate": None, "section": _getConfigRecursive(node)}
                attributeList = node.attributes                    
                for attribute in attributeList.keys():
                    if attribute in section:
                        attributeNode = attributeList.get(attribute)
                        attributeValue = attributeNode.nodeValue                        
                        section[attribute] = _removeComment(attributeValue.encode("utf-8"))

                if section["headfile"] is not None:
                    try:
                        fileHandler = open(section["headfile"], "r")
                        lineList = fileHandler.read()
                        fileHandler.close()
                        section["head"] = _removeComment(lineList)                        
                    except:
                        if __name__ == "__main__":
                            _displayMessage("Error when reading section head file '%s'." % section["headfile"], True)
                    
                if section["tailfile"] is not None:
                    try:
                        fileHandler = open(section["tailfile"], "r")
                        lineList = fileHandler.read()
                        fileHandler.close()
                        section["tail"] = _removeComment(lineList)                        
                    except:
                        if __name__ == "__main__":
                            _displayMessage("Error when reading section tail file '%s'." % section["tailfile"], True)

                if section["generatefile"] is not None:
                    try:
                        fileHandler = open(section["generatefile"], "r")
                        lineList = fileHandler.read()
                        fileHandler.close()
                        section["generate"] = _removeComment(lineList)                        
                    except:
                        if __name__ == "__main__":
                            _displayMessage("Error when reading section generate file '%s'." % section["generatefile"], True)
                    
                if section["defaultfile"] is not None:
                    try:
                        fileHandler = open(section["defaultfile"], "r")
                        lineList = fileHandler.read()
                        fileHandler.close()
                        section["default"] = _removeComment(lineList)                        
                    except:
                        if __name__ == "__main__":
                            _displayMessage("Error when reading section default file '%s'." % section["defaultfile"], True)
                        
                result.append(section)
                
    
    return result




def getData(pInputList, pKeywordList = {}, pFile = None, pType = None, pParameter = None, pRange = None):
    """
    Description:   
       This function reads data from the file specified in the parameter "pFile" using the plug-in specified in parameter "pType"
       to accomplish such task. The parameter "pParameter" is passed to the plug-in in order for it to have additional information
       on how to handle the input file. The parameter "pRange" selects the data of interest. Note that if one of these four parameters
       is not defined, then its value will be retrieved from the parameter "pInputList". The parameter "pInputList" contains a
       structure indicating what data to read from the input file and is passed to the plug-in. The function returns three elements:
       code, data and message.           

    Parameter(s):
       pInputList - list containing a structure indicating what data to read from the input file.
       pKeywordList - contains a list of keywords to be replaced in the input. If not provided, this parameter is set to empty (i.e. {}).       
       pFile - string containing the name of the input file to be read. If not provided, this parameter will be retrieved from the "pInputList".
       pType - string containing the type of the input file (basically, the plug-in to use for reading the data). If not provided, this parameter will be retrieved from the "pInputList".
       pParameter - string containing additional information that will be passed to the plug-in order to know how to handle the input file. If not provided, this parameter will be retrieved from the "pInputList".
       pRange - string containing the range of data of interest. If not provided, this parameter will be retrieved from the "pInputList".       
    
    Return(s):
       int - 0 (the file was successfully read) or -1 (error found when reading the file).      
       list - contains the data that was read from the input file (this list will be empty if some error was found).
       list - contains a list of messages that were generated during the reading of data.  
    
    Example:
       (...)
       code, dataList, messageList = getData(inputList, "myInput.xls")   # read data from an EXCEL file called "myInput.xls"
       
       if code == 0:
          for data in dataList:
             print data   # print each element of the data read
       else:
          print "Error found when reading input file."

       for message in messageList:
          print message   # print messages that were generated during the reading of data                    
       (...)    
    """  
    # =============================================
    #  CHECK IF INPUT FILE & TYPE ARE DEFINED
    # =============================================
    if pFile is None:    
        file = pInputList["file"]
    else:
        file = pFile                                
    if pType is None:    
        type = pInputList["type"]
    else:
        type = pType
    if pParameter is None:    
        parameter = pInputList["parameter"]
    else:
        parameter = pParameter    
    if pRange is None:    
        inputRange = pInputList["range"]
    else:
        inputRange = pRange       
    pluginList = []
    dataList = []
    for source in pInputList["source"]:
        sourceFile = _replaceKeyword(source["file"], pKeywordList, False)
        if sourceFile is None:
            tmpFile0 = file
        else:
            tmpFile0 = sourceFile
        if tmpFile0 is None:
            return -1, [], ["Please specify the input file (example: -inputfile:/users/dummy/myInput.xls)."]                           
        sourceType = source["type"]
        if sourceType is None:
            tmpType0 = type
        else:
            tmpType0 = sourceType
        if tmpType0 is None:
            return -1, [], ["The input file '%s' has no type defined." % tmpFile0]
        if source["parameter"] is None:
            source["parameter"] = parameter        
        flag = True
        for tmpFile1, tmpType1, positionlist, sourceList in pluginList:
            if tmpFile0 == tmpFile1 and tmpType0 == tmpType1:                                
                positionlist.append(len(dataList))
                sourceList.append(source)
                flag = False
                break
        if flag:
            pluginList.append([tmpFile0, tmpType0, [len(dataList)], [source]])
        dataList.append([])        
        

    # =============================================
    #  READ DATA USING THE APPROPRIATE PLUG-IN
    # =============================================                    
    systemPath = sys.path 
    directoryName = os.path.dirname(sys.argv[0])
    if directoryName == "":
        sys.path.append(".")
    else:
        sys.path.extend([directoryName, "."])       
    
    
    lastType = ""
    messageList = []
    for file, type, positionList, sourceList in pluginList:        
        
        # =============================================
        #  IMPORT PLUG-IN (ONLY IF THE CURRENT ONE IS DIFFERENT FROM THE LAST ONE)
        # =============================================                            
        if type != lastType:
            module = "Flex%s" % type
            try:       
                plugin = __import__(module)
            except:
                sys.path = systemPath
                return -1, [], ["The plugin '%s.py' was not found." % module]            
            if "getData" not in dir(plugin):
                sys.path = systemPath
                return -1, [], ["The plugin '%s.py' doesn't have a function called 'getData'." % module]
            lastType = type
        
        
        # =============================================
        #  CALL FUNCTION "getData" FROM PLUG-IN
        # =============================================
        message = "Reading input file '%s' using plugin '%s.py'." % (file, module)
        if __name__ == "__main__":
            _displayMessage(message)
        else:                                                                        
            messageList.append(message)
        try:
            code, tmpHeaderList, tmpDataList, tmpMessageList = plugin.getData(file, sourceList)
        except Exception, e:
            if isinstance(e, TypeError) or isinstance(e, ValueError):
                return -1, [], ["The function 'getData' in plugin '%s.py' should accept two parameters (file, source) and return four values (code, header, data, message)." % module]
            else:
                return -1, [], ["An unexpected error was found in function 'getData' of plugin '%s.py'." % module, errorLogger()]
        if code.__class__ != types.IntType:
            return -1, [], ["The first value returned from function 'getData' in plugin '%s.py' is not an integer (code)." % module]
        if tmpHeaderList.__class__ != types.ListType:
            return -1, [], ["The second value returned from function 'getData' in plugin '%s.py' is not a list (header)." % module]                   
        if tmpDataList.__class__ != types.ListType:
            return -1, [], ["The third value returned from function 'getData' in plugin '%s.py' is not a list (data)." % module]            
        if tmpMessageList.__class__ != types.ListType:
            return -1, [], ["The fourth value returned from function 'getData' in plugin '%s.py' is not a list (message)." % module]
        
        messageList.extend(tmpMessageList)                   
        if code == -1:
            sys.path = systemPath
            return -1, tmpDataList, messageList
        
        
        # =============================================
        #  SETUP AND SHOW HEADER IF SPECIFIED
        # =============================================
        showAutoSelectFlag = globals()["_SHOWAUTOSELECT_FLAG"]
        for i in range(len(sourceList)):   
            if sourceList[i]["autoselect"] is not None:
                if tmpHeaderList == [] or i > len(tmpHeaderList) - 1 or tmpHeaderList[i] == []:
                    message = "The source '%s' has the autoselect enabled but no header was returned." % sourceList[i]["name"]
                    if __name__ == "__main__":
                        _displayMessage(message, True)
                    else:                                                                        
                        messageList.append(message)
                else:
                    if showAutoSelectFlag:
                        if sourceList[i]["name"] is None:
                            print "* BEGIN HEADER OF SOURCE #%s *" % (i + 1)
                        else:
                            print "* BEGIN HEADER OF SOURCE '%s' *" % sourceList[i]["name"]
                    for j in range(len(tmpHeaderList[i])):
                        header = tmpHeaderList[i][j].encode("utf-8")
                        sourceList[i]["data"].insert(j, {"name": header, "select": None, "regexpselect": None, "default": None, "split": None, "match": None, "nomatch": None, "replace": None, "multiplereplace": None, "substring": None, "type": _STRING})
                        if showAutoSelectFlag and header != "":
                            print header
                    if showAutoSelectFlag:
                        if sourceList[i]["name"] is None:
                            print "* END HEADER OF SOURCE #%s *" % (i + 1)
                        else:                        
                            print "* END HEADER OF SOURCE '%s' *" % sourceList[i]["name"]
                        

        # =============================================
        #  SETUP DATA
        # =============================================        
        try:

            for i in range(len(positionList)):
                dataList[positionList[i]] = tmpDataList[i]
        except:
            sys.path = systemPath
            messageList.append("The plugin '%s.py' returned data with an unexpected structure." % module)
            messageList.append(errorLogger())
            return -1, [], messageList
        

    sys.path = systemPath    

   
    # =============================================
    #  PROCESS DATA
    # =============================================         
    try:       
        # =============================================
        #  PREPARE KEYWORDS LIST FOR GETDATA
        #  FOR ALL SOURCES (if combine == TRUE)
        # =============================================
        if pInputList["combine"] is not None and pInputList["combine"].upper() == "TRUE":
            if pKeywordList["inputfile"][0] is not None:
                inputFile = pKeywordList["inputfile"][0]   
            else:
                inputFile = ""
            allSourceFiles = ""
            allSourceNames = ""
            allSourceAlias = ""
            referenceList = {}
            i = len(pInputList["source"])
    
            for j in range(len(pInputList["source"])):
                if pInputList["source"][j]["file"] is None:
                    if pInputList["file"] is None:
                        sourceFile = inputFile
                    else:
                        sourceFile = pInputList["file"]
                else:
                    sourceFile = pInputList["source"][j]["file"]
                sourceName = pInputList["source"][j]["name"]
                sourceAlias = pInputList["source"][j]["alias"]
                dataListTemp = pInputList["source"][j]["data"]
                rowCount = len(dataList[j])
                if allSourceFiles == "":                
                    allSourceFiles = sourceFile
                    allSourceNames = sourceName
                    allRowCount = rowCount
                    if sourceAlias is None:
                        allSourceAlias = ""
                    else:
                        allSourceAlias = sourceAlias
                else:
                    allSourceFiles = "%s,%s" % (allSourceFiles, sourceFile)
                    allSourceNames = "%s,%s" % (allSourceNames, sourceName)
                    allRowCount = "%s,%s" % (allRowCount, rowCount)
                    if sourceAlias is None:
                        allSourceAlias = "%s," % allSourceAlias
                    else:
                        allSourceAlias = "%s,%s" % (allSourceAlias, sourceAlias)
                pKeywordList["inputfile.%s" % (j + 1)] = [sourceFile, _STRING]
                pKeywordList["sourcename.%s" % (j + 1)] = [sourceName, _STRING]
                if sourceAlias is None:
                    pKeywordList["sourcealias.%s" % (j + 1)] = ["", _STRING]
                else:
                    pKeywordList["sourcealias.%s" % (j + 1)] = [sourceAlias, _STRING]                
                pKeywordList["rowcount.%s" % (j + 1)] = [rowCount, _INTEGER]            
                for data in dataListTemp:
                    reference = [i, data["type"]]
                    referenceList["%s.%s" % (sourceName, data["name"])] = reference
                    if sourceAlias is not None:
                        referenceList["%s.%s" % (sourceAlias, data["name"])] = reference
                    i += 1
    
            pKeywordList["inputfile"] = pKeywordList["inputfile.0"] = [allSourceFiles, _STRING]     
            pKeywordList["sourcename"] = pKeywordList["sourcename.0"] = [allSourceNames, _STRING]
            pKeywordList["sourcealias"] = pKeywordList["sourcealias.0"] = [allSourceAlias, _STRING]
            pKeywordList["rowcount"] = pKeywordList["rowcount.0"] = [allRowCount, _STRING]  

        
        # =============================================
        #  CREATE REFERENCE LIST FOR THE "condition" IN
        #  THE "source" ELEMENT
        # =============================================
        combineFlag = pInputList["combine"] is not None and pInputList["combine"].upper() == "TRUE"
        referenceList = []
        for source0 in pInputList["source"]:
            sourceName = source0["name"]
            sourceAlias = source0["alias"]
            tmpDataList = source0["data"]
            tmpReferenceList = {}            
            i = 1
            for data in tmpDataList:                    
                reference = [i, data["type"]]                 
                tmpReferenceList["%s" % data["name"]] = reference
                if sourceName is not None:                
                    tmpReferenceList["%s.%s" % (sourceName, data["name"])] = reference
                if sourceAlias is not None:
                    tmpReferenceList["%s.%s" % (sourceAlias, data["name"])] = reference
                i += 1
            referenceList.append(tmpReferenceList)
            
        
        result = []               
        for i in range(len(dataList)):   # for each source            
            # =============================================
            #  PREPARE KEYWORDS LIST FOR GETDATA FOR 
            #  CURRENT SOURCE (if combine is None or False)
            # =============================================           
            if not (pInputList["combine"] is not None and pInputList["combine"].upper() == "TRUE"):
                if pKeywordList["inputfile"][0] is not None:
                    inputFile = pKeywordList["inputfile"][0]   
                else:
                    inputFile = ""           
                allSourceFiles = ""
                allSourceNames = ""
                allSourceAlias = ""
    
                if pInputList["source"][i]["file"] is None:
                    if pInputList["file"] is None:
                        sourceFile = inputFile
                    else:
                        sourceFile = pInputList["file"]
                else:
                    sourceFile = pInputList["source"][i]["file"]
                sourceName = pInputList["source"][i]["name"]
                sourceAlias = pInputList["source"][i]["alias"]
                rowCount = len(dataList[i])
                allSourceFiles = sourceFile
                allSourceNames = sourceName
                allRowCount = rowCount
                if sourceAlias is None:
                    allSourceAlias = ""
                else:
                    allSourceAlias = sourceAlias
                pKeywordList["inputfile"] = pKeywordList["inputfile.0"] = [allSourceFiles, _STRING]     
                pKeywordList["sourcename"] = pKeywordList["sourcename.0"] = [allSourceNames, _STRING]
                pKeywordList["sourcealias"] = pKeywordList["sourcealias.0"] = [allSourceAlias, _STRING]
                pKeywordList["rowcount"] = pKeywordList["rowcount.0"] = [allRowCount, _STRING]               
            
            sourceList = pInputList["source"][i]
            
            # =============================================
            #  INSERT ROW INDEX
            # =============================================
            for j in range(len(dataList[i])):    # for each row
                dataList[i][j].insert(0, j + 1)
                                

            # =============================================
            #  APPLY SOURCE "range" IF SPECIFIED
            # =============================================
            if sourceList["range"] is None:
                dataRange, tmpMessageList = _applyRange(dataList[i], inputRange)
            else:
                dataRange, tmpMessageList = _applyRange(dataList[i], sourceList["range"])                
            if __name__ == "__main__":
                for message in tmpMessageList:
                    _displayMessage(message)
            else:                                                                        
                messageList.extend(tmpMessageList)                


            # =============================================
            #  APPLY SOURCE "unique" IF SPECIFIED
            # =============================================     
            if sourceList["unique"] is not None and sourceList["unique"].upper() == "TRUE":
                uniqueList = []
                for line0 in dataRange:
                    line1 = line0[1:]
                    flag = True
                    for line2 in uniqueList:
                        if line1 == line2[1:]:
                            flag = False
                            break
                    if flag:
                        uniqueList.append(line0)
                dataRange = uniqueList
            

            # =============================================
            #  APPLY DATA "default" IF SPECIFIED (ONLY TEXT)
            # =============================================            
            j = 0
            while(j < len(dataRange)):    # for each row                
                k = 1                
                while(k < len(dataRange[j])):   # for each column
                    if dataRange[j][k] is None:
                        if sourceList["data"][k - 1]["default"] is None:
                            dataRange[j][k] = ""
                    else:
                        try:
                            dataRange[j][k] = dataRange[j][k].encode("utf-8")
                        except:
                            dataRange[j][k] = str(dataRange[j][k])
                    k += 1
                j += 1
                                

            j = 0
            while(j < len(dataRange)):    # for each row                
                k = 1
                while(k < len(dataRange[j])):   # for each column
                    if dataRange[j][k] is not None:                    
                        # =============================================
                        #  APPLY DATA "split" IF SPECIFIED
                        # =============================================
                        dataSplit = sourceList["data"][k - 1]["split"]
                        if dataSplit is not None:
                            dataRangeSplit = dataRange[j][k].split(dataSplit)
                            m = len(dataRangeSplit)
                            if m > 1:
                                for n in range(1, m):
                                    tmpDataRange = []
                                    tmpDataRange.extend(dataRange[j])
                                    dataRange.insert(j + n, tmpDataRange)
                                for n in range(m):
                                    dataRange[j + n][k] = dataRangeSplit[n]                                
    
                        
                        # =============================================
                        #  APPLY DATA "match" IF SPECIFIED
                        # =============================================
                        sourceMatch = sourceList["data"][k - 1]["match"]
                        if sourceMatch is not None:
                            sourceMatch, tmpMessageList = _applyCompute(_replaceKeyword(sourceMatch, pKeywordList, False))
                            if __name__ == "__main__":                        
                                for message in tmpMessageList:
                                    _displayMessage(message)
                            else:                                                                        
                                messageList.extend(tmpMessageList)                            
                            if re.search(sourceMatch, dataRange[j][k]) is None:
                                del dataRange[j]
                                j -= 1
                                break



                        # =============================================
                        #  APPLY DATA "nomatch" IF SPECIFIED
                        # =============================================
                        sourceNoMatch = sourceList["data"][k - 1]["nomatch"]
                        if sourceNoMatch is not None:
                            sourceNoMatch, tmpMessageList = _applyCompute(_replaceKeyword(sourceNoMatch, pKeywordList, False))
                            if __name__ == "__main__":                        
                                for message in tmpMessageList:
                                    _displayMessage(message)
                            else:                                                                        
                                messageList.extend(tmpMessageList)                            
                            if re.search(sourceNoMatch, dataRange[j][k]) is not None:
                                del dataRange[j]
                                j -= 1
                                break
                                

    
                        # =============================================
                        #  APPLY DATA "replace" IF SPECIFIED
                        # =============================================
                        dataRange[j][k], tmpMessageList = _applyReplace(dataRange[j][k], _replaceKeyword(sourceList["data"][k - 1]["replace"], pKeywordList, False), sourceList["data"][k - 1]["multiplereplace"], referenceList[i], dataRange[j])
                        if __name__ == "__main__":                        
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)                        
                        

                            
                            
                        # =============================================
                        #  APPLY DATA "substring" IF SPECIFIED
                        # =============================================
                        dataRange[j][k], tmpMessageList = _applySubString(dataRange[j][k], sourceList["data"][k - 1]["substring"])
                        if __name__ == "__main__":                        
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)
                            
                    k += 1                    
                j += 1                


            # =============================================
            #  APPLY DATA "default" IF SPECIFIED (REPLACE REFERENCES)
            # =============================================
            j = 0
            while(j < len(dataRange)):    # for each row                
                k = 1                
                while(k < len(dataRange[j])):   # for each column
                    if dataRange[j][k] is None:

                        sourceList["data"][k - 1]["default"] = _replaceKeyword(sourceList["data"][k - 1]["default"], pKeywordList, False)
                        dataRange[j][k], tmpMessageList = _applyCompute(_replaceReference(sourceList["data"][k - 1]["default"], referenceList[i], dataRange[j], False))
                        if __name__ == "__main__":
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)
                        
                        
                        # =============================================
                        #  APPLY DATA "split" IF SPECIFIED
                        # =============================================
                        dataSplit = sourceList["data"][k - 1]["split"]
                        if dataSplit is not None:
                            dataRangeSplit = dataRange[j][k].split(dataSplit)
                            m = len(dataRangeSplit)
                            if m > 1:
                                for n in range(1, m):
                                    tmpDataRange = []
                                    tmpDataRange.extend(dataRange[j])
                                    dataRange.insert(j + n, tmpDataRange)
                                for n in range(m):
                                    dataRange[j + n][k] = dataRangeSplit[n]                                
    
                        
                        # =============================================
                        #  APPLY DATA "match" IF SPECIFIED
                        # =============================================
                        sourceMatch = sourceList["data"][k - 1]["match"]
                        if sourceMatch is not None:
                            sourceMatch, tmpMessageList = _applyCompute(_replaceKeyword(sourceMatch, pKeywordList, False))
                            if __name__ == "__main__":                        
                                for message in tmpMessageList:
                                    _displayMessage(message)
                            else:                                                                        
                                messageList.extend(tmpMessageList)                            
                            if re.search(sourceMatch, dataRange[j][k]) is None:
                                del dataRange[j]
                                j -= 1
                                break
                            

                            
                        # =============================================
                        #  APPLY DATA "nomatch" IF SPECIFIED
                        # =============================================
                        sourceNoMatch = sourceList["data"][k - 1]["nomatch"]
                        if sourceNoMatch is not None:
                            sourceNoMatch, tmpMessageList = _applyCompute(_replaceKeyword(sourceNoMatch, pKeywordList, False))
                            if __name__ == "__main__":                        
                                for message in tmpMessageList:
                                    _displayMessage(message)
                            else:                                                                        
                                messageList.extend(tmpMessageList)                            
                            if re.search(sourceNoMatch, dataRange[j][k]) is not None:
                                del dataRange[j]
                                j -= 1
                                break
                                                    
    
    
                        # =============================================
                        #  APPLY DATA "replace" IF SPECIFIED
                        # =============================================
                        dataRange[j][k], tmpMessageList = _applyReplace(dataRange[j][k], _replaceKeyword(sourceList["data"][k - 1]["replace"], pKeywordList, False), sourceList["data"][k - 1]["multiplereplace"], referenceList[i], dataRange[j])
                        if __name__ == "__main__":                        
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)                        
                        
                            
                            
                        # =============================================
                        #  APPLY DATA "substring" IF SPECIFIED
                        # =============================================
                        dataRange[j][k], tmpMessageList = _applySubString(dataRange[j][k], sourceList["data"][k - 1]["substring"])
                        if __name__ == "__main__":                        
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)
                        
                    k += 1
                j += 1                
                
            
            tmpResult0 = []                                                        
            j = 0
            while(j < len(dataRange)):    # for each row
                tmpResult1 = [dataRange[j][0]]
                k = 1
                while(k < len(dataRange[j])):   # for each column                    
                    
                    # =============================================
                    #  APPLY DATA "type" CONVERSION
                    # =============================================                                              
                    type = sourceList["data"][k - 1]["type"]
                    if type == _INTEGER:
                        if dataRange[j][k] == "":
                            dataRange[j][k] = 0
                        else:
                            try:                            
                                dataRange[j][k] = int(dataRange[j][k])
                            except:
                                message = "Error when converting '%s' into an integer." % dataRange[j][k]
                                if __name__ == "__main__":
                                    _displayMessage(message)
                                else:                                                                        
                                    messageList.append(message)                                
                    elif type == _FLOAT:
                        if dataRange[j][k] == "":
                            dataRange[j][k] = 0.0
                        else:                        
                            try:                            
                                dataRange[j][k] = float(dataRange[j][k])
                            except:
                                message = "Error when converting '%s' into a float." % dataRange[j][k]
                                if __name__ == "__main__":
                                    _displayMessage(message)
                                else:                                                                        
                                    messageList.append(message)                                                                
                    tmpResult1.append(dataRange[j][k])
                    
                    k += 1
                    
                                
                # =============================================
                #  APPLY SOURCE "condition" IF SPECIFIED
                # =============================================
                condition = sourceList["condition"]
                if condition in (None, ""):
                    tmpResult0.append(tmpResult1)
                else:
                    print tmpResult1
                    condition = _replaceReference(condition, referenceList[i], tmpResult1, True)                        
                    try:
                        if eval(condition):
                            tmpResult0.append(tmpResult1)
                    except:
                        message = "Error when evaluating condition '%s'." % condition
                        if __name__ == "__main__":
                            _displayMessage(message)
                        else:                                                                        
                            messageList.append(message)                            
                j += 1


            # =============================================
            #  APPLY SOURCE "order" IF SPECIFIED
            # =============================================
            if sourceList["order"] is not None:
                sourceOrderList0 = sourceList["order"].split(":")
                sourceOrderList1 = sourceOrderList0[0].split(",")
                index = ""                                
                for sourceOrder in sourceOrderList1:
                    sourceOrder = sourceOrder.strip()                    
                    if sourceOrder[0] == "{" and sourceOrder[-1] == "}":
                        try:
                            if index == "":
                                index = referenceList[i][sourceOrder[1:-1]][0]
                            else:
                                index = "%s,%s" % (index, referenceList[i][sourceOrder[1:-1]][0])                                
                        except:
                            message = "The source '%s' doesn't have a '%s' to order." % (sourceList["name"], sourceOrder[1:-1])
                            if __name__ == "__main__":
                                _displayMessage(message)
                            else:                                                                        
                                messageList.append(message)                
                if index != "":
                    if len(sourceOrderList0) == 1 or sourceOrderList0[1].upper() == "ASCEND":
                        exec("tmpResult0 = sorted(tmpResult0, key = itemgetter(%s))" % index)
                    else:
                        if sourceOrderList0[1].upper() == "DESCEND": 
                            exec("tmpResult0 = sorted(tmpResult0, key = itemgetter(%s), reverse = True)" % index)
                
            
            result.append(tmpResult0)                                        
            
        return code, result, messageList    
    
    except KeyboardInterrupt:
        messageList.append("Processing interrupted by the user.")
        return -1, [], messageList
    except:
        messageList.append("The plugin '%s.py' returned data with an unexpected structure." % module)
        messageList.append(errorLogger())
        return -1, [], messageList
            



def generateOutput(pInputList, pOutputList, pDataList, pKeywordList = {}, pFile = None, pType = None, pParameter = None, pAppendFlag = None, pHeadList = None, pTailList = None):   
    """
    Description:
       This function transforms the data stored in the parameter "pDataList" using the "pInputList" parameter and generates the output
       using the "pOutputList" parameter to perform this task. Note that if the "pHeadList" and "pTailList" parameters are specified
       (i.e. not None), then their contents will be inserted in the beginning and in the end of the result of the output, respectively.
       A list is returned with the result of the output (i.e. the transformation of data).        
    
    Parameter(s):
       pInputList - contains a list with the input structure that was read from the XML config.
       pOutputList - contains a list with the output structure that was read from the XML config.
       pDataList - contains a list with the data that was read from the input file.
       pKeywordList - contains a list of keywords to be replaced in the output. If not provided, this parameter is set to empty (i.e. {}).
       pFile - string containing the name of the output file to be read. If not provided, this parameter will be retrieved from the "pOutputList".
       pType - string containing the type of the output file (basically, the plug-in to use for writing the data). If not provided, this parameter
            will be retrieved from the "pOutputList".
       pParameter - string containing additional information that will be passed to the plug-in order to know how to handle the output file. If not
                    provided, this parameter will be retrieved from the "pOutputList".       
       pAppendFlag - boolean that specifies if the output file is to be appended (True) or overwritten (False) with the new data. If
                     not provided, this parameter is set to None.       
       pHeadList - contains a list with data to be inserted in the beginning of the result of the output. If the list is None,
                   then the data to be inserted is the one specified in the XML config.
       pTailList - contains a list with data to be inserted in the end of the result of the output. If the list is None,
                   then the data to be inserted is the one specified in the XML config.       
    
    Return(s):
       int - 0 (the output was successfully generated) or -1 (error found when generating the output).    
       list - contain a list with the result of the output (i.e. the transformation of data).
       list - contains a list of messages that were generated during the generation of the output.
    
    Example:
       (...)
       code, resultList, messageList = generateOutput(inputList, outputList, dataList)   # transform the data
       
       if code == 0:
          for result in resultList:
             print result   # print each element of the result of the output

          for message in messageList:
             print message   # print messages that were generated during the generation of the output
       else:
          print "Error found when generating the output."
       (...)    
       
    """
    
    resultList = []
    messageList = []
    headList = {}
    headIndex = 0       
    if pInputList["combine"] is not None and pInputList["combine"].upper() == "TRUE":
        
        if pKeywordList["inputfile"][0] is not None:
            inputFile = pKeywordList["inputfile"][0]
        else:
            inputFile = ""                                                          
        allSourceFiles = ""
        allSourceNames = ""
        allSourceAlias = ""
        referenceList = {}
        i = len(pInputList["source"])
        for j in range(len(pInputList["source"])):
            if pInputList["source"][j]["file"] is None:
                if pInputList["file"] is None:
                    sourceFile = inputFile
                else:
                    sourceFile = pInputList["file"]
            else:
                sourceFile = pInputList["source"][j]["file"]
            sourceName = pInputList["source"][j]["name"]
            sourceAlias = pInputList["source"][j]["alias"]
            dataList = pInputList["source"][j]["data"]
            rowCount = len(pDataList[j])
            if allSourceFiles == "":                
                allSourceFiles = sourceFile
                allSourceNames = sourceName
                allRowCount = rowCount
                if sourceAlias is None:
                    allSourceAlias = ""
                else:
                    allSourceAlias = sourceAlias
            else:
                allSourceFiles = "%s,%s" % (allSourceFiles, sourceFile)
                allSourceNames = "%s,%s" % (allSourceNames, sourceName)
                allRowCount = "%s,%s" % (allRowCount, rowCount)
                if sourceAlias is None:
                    allSourceAlias = "%s," % allSourceAlias
                else:
                    allSourceAlias = "%s,%s" % (allSourceAlias, sourceAlias)
            pKeywordList["inputfile.%s" % (j + 1)] = [sourceFile, _STRING]
            pKeywordList["sourcename.%s" % (j + 1)] = [sourceName, _STRING]
            if sourceAlias is None:
                pKeywordList["sourcealias.%s" % (j + 1)] = ["", _STRING]
            else:
                pKeywordList["sourcealias.%s" % (j + 1)] = [sourceAlias, _STRING]                
            pKeywordList["rowcount.%s" % (j + 1)] = [rowCount, _INTEGER]            
            for data in dataList:
                reference = [i, data["type"]]
                referenceList["%s.%s" % (sourceName, data["name"])] = reference
                if sourceAlias is not None:
                    referenceList["%s.%s" % (sourceAlias, data["name"])] = reference
                i += 1

        pKeywordList["inputfile"] = pKeywordList["inputfile.0"] = [allSourceFiles, _STRING]     
        pKeywordList["sourcename"] = pKeywordList["sourcename.0"] = [allSourceNames, _STRING]
        pKeywordList["sourcealias"] = pKeywordList["sourcealias.0"] = [allSourceAlias, _STRING]
        pKeywordList["rowcount"] = pKeywordList["rowcount.0"] = [allRowCount, _STRING]        
        
        _combineData(pInputList, pOutputList, pDataList, pKeywordList, referenceList, headList, resultList, messageList, 0, 0, "", [])
        
    else:        
        
        _sequenceData(pInputList, pOutputList, pDataList, pKeywordList, headList, resultList, messageList)      
        if pKeywordList["inputfile"][0] is not None:
            inputFile = pKeywordList["inputfile"][0] 
        else:
            inputFile = ""                  
        allSourceFiles = ""
        allSourceNames = ""
        allSourceAlias = ""
        for j in range(len(pInputList["source"])):
            if pInputList["source"][j]["file"] is None:
                if pInputList["file"] is None:
                    sourceFile = inputFile
                else:
                    sourceFile = pInputList["file"]
            else:
                sourceFile = pInputList["source"][j]["file"]
            sourceName = pInputList["source"][j]["name"]
            sourceAlias = pInputList["source"][j]["alias"]
            dataList = pInputList["source"][j]["data"]
            rowCount = len(pDataList[j])
            if allSourceFiles == "":                
                allSourceFiles = sourceFile
                allSourceNames = sourceName
                allRowCount = rowCount
                if sourceAlias is None:
                    allSourceAlias = ""
                else:
                    allSourceAlias = sourceAlias
            else:
                allSourceFiles = "%s,%s" % (allSourceFiles, sourceFile)
                allSourceNames = "%s,%s" % (allSourceNames, sourceName)
                allRowCount = "%s,%s" % (allRowCount, rowCount)
                if sourceAlias is None:
                    allSourceAlias = "%s," % allSourceAlias
                else:
                    allSourceAlias = "%s,%s" % (allSourceAlias, sourceAlias)
            pKeywordList["inputfile.%s" % (j + 1)] = [sourceFile, _STRING]
            pKeywordList["sourcename.%s" % (j + 1)] = [sourceName, _STRING]
            if sourceAlias is None:
                pKeywordList["sourcealias.%s" % (j + 1)] = ["", _STRING]
            else:
                pKeywordList["sourcealias.%s" % (j + 1)] = [sourceAlias, _STRING]                
            pKeywordList["rowcount.%s" % (j + 1)] = [rowCount, _INTEGER]            
    
        pKeywordList["inputfile"] = pKeywordList["inputfile.0"] = [allSourceFiles, _STRING]     
        pKeywordList["sourcename"] = pKeywordList["sourcename.0"] = [allSourceNames, _STRING]
        pKeywordList["sourcealias"] = pKeywordList["sourcealias.0"] = [allSourceAlias, _STRING]
        pKeywordList["rowcount"] = pKeywordList["rowcount.0"] = [allRowCount, _STRING]        

    # =============================================
    #  REMOVE SPECIAL KEYWORD "rowindex" FROM LIST
    #  SINCE IT SHOULDN'T BE REFERENCED
    # =============================================
    for key in pKeywordList.keys():
        if key.startswith("rowindex"):
            del pKeywordList[key]


    # =============================================
    #  APPLY SECTION "orderhead" IF SPECIFIED
    # =============================================       
    index = 1
    flag0 = True
    while(flag0):
        flag0 = False
        for i in range(len(resultList) - 1):
            if resultList[i]["index"] == index:
                if not flag0:
                    flag0 = True
                    if resultList[i]["orderhead"] is None:
                        flag1 = -1
                    else:
                        if resultList[i]["orderhead"].upper() == "ASCEND":
                            flag1 = 0
                        elif resultList[i]["orderhead"].upper() == "DESCEND":
                            flag1 = 1
                        else:
                            flag1 = -1
                if flag1 != -1:
                    for j in range(i + 1, len(resultList)):
                        if resultList[j]["index"] == index:
                            if flag1 == 0:
                                if resultList[i]["head"] > resultList[j]["head"]:
                                    tmpResultList = resultList[i]
                                    resultList[i] = resultList[j]
                                    resultList[j] = tmpResultList
                            else:
                                if resultList[i]["head"] < resultList[j]["head"]:
                                    tmpResultList = resultList[i]
                                    resultList[i] = resultList[j]
                                    resultList[j] = tmpResultList
        index += 1


    
    # =============================================
    #  GENERATE A SINGLE LIST FROM ALL SECTIONS RESULTS
    # =============================================
    result, tmpMessageList = _generateAllOutput(resultList, None)
    messageList.extend(tmpMessageList)
    

    outputSplit = pOutputList["split"]
    outputMatch = pOutputList["match"]
    outputNoMatch = pOutputList["nomatch"]
    outputReplace = pOutputList["replace"]
    outputSubString = pOutputList["substring"]
    i = 0
    while(i < len(result)):
        
        # =============================================
        #  APPLY OUTPUT "condition" IF SPECIFIED
        # =============================================       
        tmpCondition = _replaceKeyword(pOutputList["condition"], pKeywordList, True)
        if tmpCondition in (None, ""):        
            conditionFlag = True
        else:
            try:
                conditionFlag = eval(tmpCondition)               
            except:
                conditionFlag = False
                message = "Error when evaluating condition '%s'." % tmpCondition
                if __name__ == "__main__":                        
                    _displayMessage(message)
                else:                                                                        
                    messageList.append(message)
                    
        if conditionFlag:       
        
            # =============================================
            #  APPLY OUTPUT "split" IF SPECIFIED
            # =============================================
            if outputSplit is not None:
                resultSplit = result[i].split(outputSplit)
                j = len(resultSplit)
                if j > 1:
                    result[i] = resultSplit[0]
                    for k in range(1, j):
                        result.insert(i + k, resultSplit[k])                            
            
        
            # =============================================
            #  APPLY OUTPUT "match" IF SPECIFIED
            # =============================================            
            if outputMatch is not None:
                conditionFlag = (re.search(outputMatch, result[i]) is not None)
                                                                  
           
            if conditionFlag:
                
                # =============================================
                #  APPLY OUTPUT "nomatch" IF SPECIFIED
                # =============================================            
                if outputNoMatch is not None:
                    conditionFlag = (re.search(outputNoMatch, result[i]) is None)
                
                              
                if conditionFlag:
                                            
                    # =============================================
                    #  APPLY OUTPUT "replace" IF SPECIFIED
                    # =============================================
                    if outputReplace is not None:
                        result[i], tmpMessageList = _applyReplace(result[i], outputReplace, pOutputList["multiplereplace"], None, None)
                        if __name__ == "__main__":                        
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)                 
                                
                                
                    # =============================================
                    #  APPLY OUTPUT "substring" IF SPECIFIED
                    # =============================================
                    if outputSubString is not None:      
                        result[i], tmpMessageList = _applySubString(result[i], outputSubString)
                        if __name__ == "__main__":                        
                            for message in tmpMessageList:
                                _displayMessage(message)
                        else:                                                                        
                            messageList.extend(tmpMessageList)
                                
                    i += 1
                    
                else:
                    del result[i]
                
            else:
                del result[i]
                
        else:
            del result[i]
                
                
    # =============================================
    #  APPLY OUTPUT "unique" IF SPECIFIED
    # =============================================     
    result, tmpMessageList = _applyUnique(result, pOutputList["unique"])
    if __name__ == "__main__":                        
        for message in tmpMessageList:
            _displayMessage(message)
    else:                                                                        
        messageList.extend(tmpMessageList)            


    # =============================================
    #  APPLY OUTPUT "range" IF SPECIFIED
    # =============================================              
    result, tmpMessageList = _applyRange(result, pOutputList["range"])
    if __name__ == "__main__":
        for message in tmpMessageList:
            _displayMessage(message)
    else:                                                                        
        messageList.extend(tmpMessageList)
   

    # =============================================
    #  APPLY OUTPUT "ordergenerate" IF SPECIFIED
    # =============================================     
    if pOutputList["ordergenerate"] is not None:
        if pOutputList["ordergenerate"].upper() == "ASCEND":
            result.sort()
        elif pOutputList["ordergenerate"].upper() == "DESCEND":
            result.sort(reverse = True)
      
            
    # =============================================
    #  APPLY OUTPUT "joingenerate" IF SPECIFIED
    # ============================================= 
    if pOutputList["joingenerate"] is not None:
        result = [pOutputList["joingenerate"].join(result)]
    
    # =============================================
    #  ADD OUTPUT HEAD AND/OR TAIL IF SPECIFIED
    # =============================================    
    if pHeadList is None:
        if pOutputList["headfile"] is not None:
            try:
                fileHandler = open(pOutputList["headfile"], "r")
                lineList = fileHandler.readlines()
                fileHandler.close()
                i = 0
                for line in lineList:
                    if line[-1] == "\n":
                        line, tmpMessageList = _applyCompute(_replaceKeyword(line[:-1], pKeywordList, False))
                    else:
                        line, tmpMessageList = _applyCompute(_replaceKeyword(line, pKeywordList, False))
                    result.insert(i, line)
                    if __name__ == "__main__":                        
                        for message in tmpMessageList:
                            _displayMessage(message)
                    else:                                                                        
                        messageList.extend(tmpMessageList)                    
                    i += 1
            except:
                messageList.append("Error when reading head file '%s'." % pOutputList["headfile"])
        else:
            if pOutputList["head"] is not None:
                line, tmpMessageList = _applyCompute(_replaceKeyword(pOutputList["head"], pKeywordList, False))
                result.insert(0, line)
                if __name__ == "__main__":                        
                    for message in tmpMessageList:
                        _displayMessage(message)
                else:                                                                        
                    messageList.extend(tmpMessageList)              
    else:
        i = 0
        for line in pHeadList:
            line, tmpMessageList = _applyCompute(_replaceKeyword(line, pKeywordList, False))
            result.insert(i, line)
            if __name__ == "__main__":                        
                for message in tmpMessageList:
                    _displayMessage(message)
            else:                                                                        
                messageList.extend(tmpMessageList)                          
            i += 1       
        
    if pTailList is None:
        if pOutputList["tailfile"] is not None:
            try:
                fileHandler = open(pOutputList["tailfile"], "r")
                lineList = fileHandler.readlines()
                fileHandler.close()
                for line in lineList:
                    if line[-1] == "\n":
                        line, tmpMessageList = _applyCompute(_replaceKeyword(line[:-1], pKeywordList, False))
                    else:
                        line, tmpMessageList = _applyCompute(_replaceKeyword(line, pKeywordList, False))
                    result.append(line)
                    if __name__ == "__main__":                        
                        for message in tmpMessageList:
                            _displayMessage(message)
                    else:                                                                        
                        messageList.extend(tmpMessageList)
            except:
                messageList.append("Error when reading tail file '%s'." % pOutputList["tailfile"])
        else:
            if pOutputList["tail"] is not None:
                line, tmpMessageList = _applyCompute(_replaceKeyword(pOutputList["tail"], pKeywordList, False))
                result.append(line)
                if __name__ == "__main__":                        
                    for message in tmpMessageList:
                        _displayMessage(message)
                else:                                                                        
                    messageList.extend(tmpMessageList)                
    else:
        for line in pTailList:
            line, tmpMessageList = _applyCompute(_replaceKeyword(line, pKeywordList, False))
            result.append(line)
            if __name__ == "__main__":                        
                for message in tmpMessageList:
                    _displayMessage(message)
            else:                                                                        
                messageList.extend(tmpMessageList)
              
    # =============================================
    #  SET DATA USING THE APPROPRIATE PLUG-IN IF SPECIFIED
    # =============================================
    if pFile is None:
        #file = pOutputList["file"]
        file = _replaceKeyword(pOutputList["file"], pKeywordList, False)
    else:
        file = pFile
    if pType is None:
        type = pOutputList["type"]
    else:
        type = pType
    if pParameter is None:
        parameter = pOutputList["parameter"]
    else:
        parameter = pParameter
    if pAppendFlag is None:
        appendFlag = pOutputList["append"] is not None and pOutputList["append"].upper() == "TRUE"
    else:
        appendFlag = pAppendFlag
        
    if file is not None and type is not None:
        systemPath = sys.path 
        directoryName = os.path.dirname(sys.argv[0])
        if directoryName == "":
            sys.path.append(".")
        else:
            sys.path.extend([directoryName, "."])      
    
        module = "Flex%s" % type
        try:        
            plugin = __import__(module)
        except:
            sys.path = systemPath
            return -1, [], ["The plugin '%s.py' was not found." % module]            
        if "setData" not in dir(plugin):
            sys.path = systemPath
            return -1, [], ["The plugin '%s.py' doesn't have a function called 'setData'." % module]    


        # =============================================
        #  CALL FUNCTION "setData" FROM PLUG-IN
        # =============================================
        if appendFlag:
            message = "Appending output to file '%s' using plugin '%s.py'." % (file, module)
        else:
            message = "Writing output to file '%s' using plugin '%s.py'." % (file, module)
        if __name__ == "__main__":
            _displayMessage(message)
        else:                                                                        
            messageList.append(message)
        try:
            code, tmpMessageList = plugin.setData(file, parameter, result, appendFlag)
        except Exception, e:
            if isinstance(e, TypeError) or isinstance(e, ValueError):
                return -1, [], ["The function 'setData' in plugin '%s.py' should accept four parameters (file, parameter, data, append) and return two values (code, message)." % module]
            else:
                return -1, [], ["An unexpected error was found in function 'setData' of plugin '%s.py'." % module, errorLogger()]      
        if code.__class__ != types.IntType:
            return -1, [], ["The first value returned from function 'setData' in plugin '%s.py' is not an integer (code)." % module]            
        if tmpMessageList.__class__ != types.ListType:
            return -1, [], ["The second value returned from function 'setData' in plugin '%s.py' is not a list (message)." % module]            
        messageList.extend(tmpMessageList)                           
        if code == -1:
            sys.path = systemPath
            return -1, [], messageList
            
        sys.path = systemPath              
    elif file is not None:
        if appendFlag:
            message = "Appending output to file '%s'." % file
        else:
            message = "Writing output to file '%s'." % file
            if __name__ == "__main__":
                _displayMessage(message)
            else:                                                                        
                messageList.append(message)              
        if _saveOutput(file, result, appendFlag) == -1:
            return -1, [], ["Error when writing output to file '%s'." % file, errorLogger()]
    else:
        for resultLine in result:
            print resultLine
     
    return 0, result, messageList


        
def _sequenceData(pInputList, pOutputList, pDataList, pKeywordList, pHeadList, pResultList, pMessageList):
    inputFile = pKeywordList["inputfile"]    
    
    for i in range(len(pDataList)):  # for each source
        sourceName = pInputList["source"][i]["name"]
        sourceAlias = pInputList["source"][i]["alias"]
        dataList = pInputList["source"][i]["data"]
        pKeywordList["sourcename"] = [sourceName, _STRING]
        if sourceAlias is None:
            pKeywordList["sourcealias"] = ["", _STRING]
        else:
            pKeywordList["sourcealias"] = [sourceAlias, _STRING]
        pKeywordList["rowcount"] = [len(pDataList[i]), _INTEGER]
        referenceList = {}
        j = 1
        for data in dataList:
            reference = [j, data["type"]]
            referenceList["%s" % data["name"]] = reference
            if sourceName is not None:
                referenceList["%s.%s" % (sourceName, data["name"])] = reference
            if sourceAlias is not None:
                referenceList["%s.%s" % (sourceAlias, data["name"])] = reference            
            j += 1
        if pInputList["source"][i]["file"] is None:
            if pInputList["file"] is None:
                pKeywordList["inputfile"] = inputFile 
            else:
                pKeywordList["inputfile"] = [pInputList["file"], _STRING]
        else:
            pKeywordList["inputfile"] = [pInputList["source"][i]["file"], _STRING]
        pKeywordList["inputfile.0"] = pKeywordList["inputfile"]
        pKeywordList["sourcename.0"] = pKeywordList["sourcename"] 
        pKeywordList["sourcealias.0"] = pKeywordList["sourcealias"]
        pKeywordList["rowcount.0"] = pKeywordList["rowcount"]
        for j in range(len(pDataList[i])):   # for each row
            pKeywordList["rowindex"] = [int(pDataList[i][j][0]), _INTEGER]
            _generateOutput(pInputList, pOutputList["section"], pKeywordList, referenceList, pHeadList, pResultList, pMessageList, pDataList[i][j], True, None)                    
            
            
            
def _combineData(pInputList, pOutputList, pDataList, pKeywordList, pReferenceList, pHeadList, pResultList, pMessageList, pDataIndex, pSourceEndCount, pRowIndex, pResult):      

    if pDataIndex < len(pDataList):
        i = len(pDataList[pDataIndex])
        for j in range(i):
            data = pDataList[pDataIndex][j]
            pKeywordList["rowindex.%s" % (pDataIndex + 1)] = [data[0], _INTEGER]
            tmpData = []
            tmpData.extend(pResult)
            tmpData.extend(data[1:])
            tmpData.insert(pDataIndex, data[0])               
            if pDataIndex == 0:
                pSourceEndCount = 0
                _initGenerateFlag(pResultList)
            else:
                if j == i - 1:
                    pSourceEndCount = pSourceEndCount + 1                    
            if pRowIndex == "":
                _combineData(pInputList, pOutputList, pDataList, pKeywordList, pReferenceList, pHeadList, pResultList, pMessageList, pDataIndex + 1, pSourceEndCount, str(data[0]), tmpData)
            else:
                _combineData(pInputList, pOutputList, pDataList, pKeywordList, pReferenceList, pHeadList, pResultList, pMessageList, pDataIndex + 1, pSourceEndCount, "%s,%s" % (pRowIndex, data[0]), tmpData)                     
    else:
        pKeywordList["rowindex"] = pKeywordList["rowindex.0"] = [pRowIndex, _STRING]
        condition = pInputList["condition"]                                  
        if condition in (None, ""):
            _generateOutput(pInputList, pOutputList["section"], pKeywordList, pReferenceList, pHeadList, pResultList, pMessageList, pResult, pSourceEndCount == len(pInputList["source"]) - 1, None)
        else:                 
            condition = _replaceReference(condition, pReferenceList, pResult, True)
            try:
                if eval(condition):
                    _generateOutput(pInputList, pOutputList["section"], pKeywordList, pReferenceList, pHeadList, pResultList, pMessageList, pResult, pSourceEndCount == len(pInputList["source"]) - 1, None)
            except:
                pMessageList.append("Error when evaluating condition '%s'." % condition)        
    


def _initGenerateFlag(pResultList):
    
    for result in pResultList:
        result["generateflag"] = False
        _initGenerateFlag(result["section"])   
       



def _generateOutput(pInputList, pOutputList, pKeywordList, pReferenceList, pHeadList, pResultList, pMessageList, pDataList, pSourceEnd, pHeadParent):
    
    combineFlag = pInputList["combine"] is None or pInputList["combine"].upper() != "TRUE"    
    
    for i in range(len(pOutputList)):        
        section = pOutputList[i]      
        defaultFlag = False
       
        # =============================================
        #  APPLY SECTION "condition" IF SPECIFIED
        # =============================================
        tmpCondition = _replaceKeyword(section["condition"], pKeywordList, True)
        if tmpCondition in (None, ""):
            processFlag = True
            conditionFlag = True
        else:
            tmpCondition = _replaceReference(tmpCondition, pReferenceList, pDataList, True)
            try:
                processFlag = True
                conditionFlag = evaluate_condition(tmpCondition)
            except:
                processFlag = False
                pMessageList.append("Error when evaluating condition '%s'." % tmpCondition)
                                                                                           
        # =============================================
        #  PROCESS
        # =============================================       
        if processFlag:            
            if conditionFlag:
                tmpGenerate = _replaceKeyword(section["generate"], pKeywordList, False)                                                        
                tmpGenerate = _replaceReference(tmpGenerate, pReferenceList, pDataList, False)         
                tmpGenerate, message = _applyCompute(tmpGenerate)
                pMessageList.extend(message)                
                generateFlag = True
            else:
                if section["default"] is None:
                    processFlag = False
                else:                    
                    defaultFlag = section["defaultunique"] is not None and section["defaultunique"].upper() == "TRUE"
                    
                    if combineFlag or section["defaultunique"] is None or section["defaultunique"].upper() != "TRUE" or (pSourceEnd and (i + 1 > len(pResultList) or not pResultList[i]["generateflag"])):                    
                        tmpDefault = _replaceKeyword(section["default"], pKeywordList, False)
                        tmpDefault = _replaceReference(tmpDefault, pReferenceList, pDataList, False)
                        tmpGenerate, message = _applyCompute(tmpDefault)
                        pMessageList.extend(message)
                        generateFlag = False
                    else:
                        processFlag = False
            
            if len(pResultList) > i and pResultList[i]["defaultunique"] and not conditionFlag:
                processFlag = False
                
            if len(pResultList) > i and defaultFlag:
                pResultList[i]["defaultunique"] = defaultFlag            
                    
            if processFlag:
                
                
                # =============================================
                #  REPLACE KEYWORDS WITH APPROPRIATE VALUES
                # =============================================
                tmpHead = _replaceKeyword(section["head"], pKeywordList, False)           
                tmpTail = _replaceKeyword(section["tail"], pKeywordList, False)
        
        
                # =============================================
                #  REPLACE REFERENCES WITH APPROPRIATE VALUES
                # =============================================
                tmpHead = _replaceReference(tmpHead, pReferenceList, pDataList, False)
                tmpTail = _replaceReference(tmpTail, pReferenceList, pDataList, False)
                        
                        
                # =============================================
                #  APPLY "compute" KEYWORD IF SPECIFIED
                # =============================================                                       
                tmpHead, message = _applyCompute(tmpHead)
                pMessageList.extend(message)
                tmpTail, message = _applyCompute(tmpTail)
                pMessageList.extend(message)            
                
                
                # =============================================
                #  APPLY SECTION "split" IF SPECIFIED
                # =============================================
                sectionSplit = section["split"]
                if sectionSplit is None:
                    generateList = [tmpGenerate]
                else:
                    generateList = tmpGenerate.split(sectionSplit)
                    
                    
                for tmpGenerate in generateList:
                                                
                    # =============================================
                    #  APPLY SECTION "match" IF SPECIFIED
                    # =============================================
                    if section["match"] is not None:
                        processFlag = (re.search(section["match"], tmpGenerate) is not None)
                                                                                    
                                            
                    if processFlag:
                        
                        
                        # =============================================
                        #  APPLY SECTION "nomatch" IF SPECIFIED
                        # =============================================
                        if section["nomatch"] is not None:
                            processFlag = (re.search(section["nomatch"], tmpGenerate) is None)
                            
                    
                        
                        if processFlag:
                            
                            # =============================================
                            #  SET HEADER
                            # =============================================
                            if pHeadParent is None:
                                if tmpHead is None:
                                    headParent = "flex_head_%s" % i
                                else:
                                    headParent = "flex_head_%s_%s" % (i, tmpHead)                
                            else:
                                if tmpHead is None:
                                    headParent = "%s*flex_head_%s" % (pHeadParent, i)
                                else:
                                    headParent = "%s*flex_head_%s_%s" % (pHeadParent, i, tmpHead)
                                
                            
                            try:
                                index = pHeadList[headParent]
                                if tmpGenerate is not None:
                                                                
                                    # =============================================
                                    #  APPLY SECTION "replace" IF SPECIFIED
                                    # =============================================                            
                                    if section["replace"] is not None:
                                        tmpGenerate, tmpMessageList = _applyReplace(tmpGenerate, section["replace"], section["multiplereplace"], pReferenceList, pDataList)
                                        pMessageList.extend(tmpMessageList)                            
                                    pResultList[index]["generate"].append(tmpGenerate)
                                    if generateFlag:                                
                                        pResultList[index]["generateflag"] = False
                                        
                            except:
                                index = len(pResultList)
                                if tmpGenerate is None:
                                    tmpGenerate = []      
                                else:
                                                                
                                    # =============================================
                                    #  APPLY SECTION "replace" IF SPECIFIED
                                    # =============================================                            
                                    if section["replace"] is not None:      
                                        tmpGenerate, tmpMessageList = _applyReplace(tmpGenerate, section["replace"], section["multiplereplace"], pReferenceList, pDataList)
                                        pMessageList.extend(tmpMessageList)
                                    tmpGenerate = [tmpGenerate]
                                    
                                pHeadList[headParent] = index
                                
                                pResultList.append({"head": tmpHead, "tail": tmpTail, "generate": tmpGenerate, "generateflag": generateFlag, "default": section["default"], "match": section["match"], "nomatch": section["nomatch"], "replace": section["replace"], "multiplereplace": section["multiplereplace"], "substring": section["substring"], "unique": section["unique"], "range": section["range"], "orderhead": section["orderhead"], "joinhead": section["joinhead"], "ordergenerate": section["ordergenerate"], "joingenerate": section["joingenerate"], "index": i + 1, "section": [], "defaultunique": defaultFlag})          
                            _generateOutput(pInputList, section["section"], pKeywordList, pReferenceList, pHeadList, pResultList[index]["section"], pMessageList, pDataList, pSourceEnd, headParent)



def evaluate_condition(tmpCondition):
    simple_string = "([\\w\\d'\_\-\\\]*)"
    pattern = re.compile(simple_string+"' == "+simple_string+"'$")
    match = pattern.match(tmpCondition)
    if match is not None:
        conditionFlag = (match.group(1) == match.group(2))
    else:
        conditionFlag = eval(tmpCondition)
    return conditionFlag

def _removeComment(pText):

    lastPosition = 0
    result = ""            
    while(True):
        i = pText.find("%-", lastPosition)
        if i == -1:
            result = "%s%s" % (result, pText[lastPosition:])
            break
        else:
            j = pText.find("-%", i + 2)
            if j == -1:
                result = "%s%s" % (result, pText[lastPosition:])
                break                        
            else:
                result = "%s%s" % (result, pText[lastPosition:i])
                lastPosition = j + 2
    return result

   


def _replaceKeyword(pReplace, pKeywordList, pApplyConversionFlag):  
    if pReplace is None:
        return None
    else:
        if not pApplyConversionFlag:
            lastPosition = 0
            result = ""           
            while(True):
                i = pReplace.find("%compute(", lastPosition)
                if i == -1:
                    result = "%s%s" % (result, pReplace[lastPosition:])
                    break
                else:
                    j = pReplace.find(")%", i + 9)
                    if j == -1:
                        result = "%s%s" % (result, pReplace[lastPosition:])
                        break
                    else:
                        result = "%s%s%s" % (result, pReplace[lastPosition:i], _replaceKeyword(pReplace[i:j + 2], pKeywordList, True))
                        lastPosition = j + 2
            pReplace = result
        lastPosition = 0
        result = ""      
        while(True):
            i = pReplace.find("%", lastPosition)
            if i == -1:
                result = "%s%s" % (result, pReplace[lastPosition:])
                break
            else:
                j = pReplace.find("%", i + 1)
                if j == -1:
                    result = "%s%s" % (result, pReplace[lastPosition:])
                    break
                else:
                    try:                   
                        keyword = pKeywordList[pReplace[i + 1:j]]                            
                        if pApplyConversionFlag and keyword[1] == _STRING:
                            result = "%s%s'%s'" %(result, pReplace[lastPosition:i], keyword[0].replace("\\", "\\\\").replace("'", "\\'"))
                        else:
                            result = "%s%s%s" %(result, pReplace[lastPosition:i], keyword[0])
                        lastPosition = j + 1
                    except:
                        result = "%s%s" % (result, pReplace[lastPosition:j])                
                        lastPosition = j
        return result
    
    

def _replaceReference(pReplace, pReferenceList, pDataList, pApplyConversionFlag):

    if pReplace is None:
        return None
    else:
        if not pApplyConversionFlag:
            lastPosition = 0
            result = ""            
            while(True):
                i = pReplace.find("%compute(", lastPosition)
                if i == -1:
                    result = "%s%s" % (result, pReplace[lastPosition:])
                    break
                else:
                    j = pReplace.find(")%", i + 9)
                    if j == -1:
                        result = "%s%s" % (result, pReplace[lastPosition:])
                        break
                    else:
                        result = "%s%s%s" % (result, pReplace[lastPosition:i], _replaceReference(pReplace[i:j + 2], pReferenceList, pDataList, True))
                        lastPosition = j + 2                   
            pReplace = result        
        lastPosition = 0
        result = ""
        while(True):
            #to handling curly brackets
            pReplace = pReplace.replace("\{","&openCurlyBracket&")
            pReplace = pReplace.replace("\}","&closeCurlyBracket&")
            i = pReplace.find("{", lastPosition)
            if i == -1:
                result = "%s%s" % (result, pReplace[lastPosition:])
                break
            else:
                j = pReplace.find("}", i + 1)
                if j == -1:
                    result = "%s%s" % (result, pReplace[lastPosition:])
                    break
                else:
                    try:
                        reference = pReferenceList[pReplace[i + 1:j]]
                        if pApplyConversionFlag and reference[1] == _STRING:
                            if pDataList[reference[0]] is None:
                                result = "%s%s''" %(result, pReplace[lastPosition:i])
                            else:
                                result = "%s%s'%s'" %(result, pReplace[lastPosition:i], pDataList[reference[0]].replace("\\", "\\\\").replace("'", "\\'"))
                        else:
                            if pDataList[reference[0]] is None:
                                result = "%s%s" %(result, pReplace[lastPosition:i])
                            else:
                                result = "%s%s%s" %(result, pReplace[lastPosition:i], pDataList[reference[0]])
                    except:
                        result = "%s%s" % (result, pReplace[lastPosition:j + 1])                
                    lastPosition = j + 1
        #to handling curly brackets
        result = result.replace("&openCurlyBracket&","{")
        result = result.replace("&closeCurlyBracket&","}")
        return result
    



def _generateAllOutput(pSectionList, pIndex):
    
    if pIndex is None:
        begin = 0
        end = len(pSectionList)
    else:
        begin = pIndex
        end = pIndex + 1
        
    messageList = []
    result = []    
    for i in range(begin, end):
        
        section = pSectionList[i]
        if section["index"] > 0:            
            
            tmpResult = section["generate"]
            
            # =============================================
            #  APPLY SECTION "substring" IF SPECIFIED
            # =============================================
            sectionSubString = section["substring"]
            if sectionSubString is not None:
                for j in range(len(tmpResult)):       
                    tmpResult[j], tmpMessageList = _applySubString(tmpResult[j], sectionSubString)
                    if __name__ == "__main__":                        
                        for message in tmpMessageList:
                            _displayMessage(message)
                    else:                                                                        
                        messageList.extend(tmpMessageList)
                  
        
            # =============================================
            #  APPLY SECTION "unique" IF SPECIFIED
            # =============================================
            tmpResult, tmpMessageList = _applyUnique(tmpResult, section["unique"])
            if __name__ == "__main__":                        
                for message in tmpMessageList:
                    _displayMessage(message)
            else:                                                                        
                messageList.extend(tmpMessageList)            
           

            
            # =============================================
            #  APPLY SECTION "ordergenerate" IF SPECIFIED
            # =============================================       
            if section["ordergenerate"] is not None:
                if section["ordergenerate"].upper() == "ASCEND":
                    tmpResult.sort()
                elif section["ordergenerate"].upper() == "DESCEND":
                    tmpResult.sort(reverse = True)
                    
    
                    
                    
            # =============================================
            #  APPLY SECTION "range" IF SPECIFIED
            # =============================================               
            tmpResult, tmpMessageList = _applyRange(tmpResult, section["range"])
            messageList.extend(tmpMessageList)     
            

                    
            # =============================================
            #  APPLY SECTION "joinhead" & "joingenerate" IF SPECIFIED
            # =============================================
            if section["joinhead"] is None:
                if section["joingenerate"] is None:
                    if section["head"] is not None:
                        result.append(section["head"])
                    result.extend(tmpResult)
                    tmpResult, tmpMessageList = _generateAllOutput(section["section"], None)
                    result.extend(tmpResult)
                    messageList.extend(tmpMessageList)
                    if section["tail"] is not None:
                        result.append(section["tail"])                    
                else:
                    if section["head"] is None:
                        line = section["joingenerate"].join(tmpResult)
                    else:
                        line = "%s%s" % (section["head"], section["joingenerate"].join(tmpResult))                      
                    tmpResult, tmpMessageList = _generateAllOutput(section["section"], None)
                    if len(tmpResult) > 0:
                        line = "%s\n%s\n" % (line, "\n".join(tmpResult))                                                                                          
                    if section["tail"] is not None:
                        line = "%s%s" % (line, section["tail"])
                    result.append(line)
                    messageList.extend(tmpMessageList)
            else:                
                if section["head"] is None:
                    line = "".join(tmpResult)
                else:                    
                    if section["joingenerate"] is None:
                        line = "%s%s" % (section["head"], "".join(tmpResult))
                    else:
                        line = "%s%s" % (section["head"], section["joingenerate"].join(tmpResult))
                tmpResult, tmpMessageList = _generateAllOutput(section["section"], None)
                messageList.extend(tmpMessageList)                
                if len(tmpResult) > 0:
                    line = "%s%s" % (line, "".join(tmpResult))                                                
                if section["tail"] is not None:
                    line = "%s%s" % (line, section["tail"])                
                for j in range(i + 1, len(pSectionList)):
                    if section["index"] == pSectionList[j]["index"]:                       
                        tmpResult, tmpMessageList = _generateAllOutput(pSectionList, j)
                        messageList.extend(tmpMessageList)
                        if len(tmpResult) > 0:
                            line = "%s%s%s" % (line, section["joinhead"], "".join(tmpResult))                                                                       
                result.append(line)                                                               
                    
            section["index"] = 0
           
    return result, messageList
        





def _saveOutput(pOutputFile, pDataList, pAppendFlag = False):
    """
    Description:    
       This function provides a convenient way of saving the data stored in parameter "pDataList" into a certain file specified
       in the "pOutputFile" parameter. The output file can be overwritten or appended with this new data, according to the value
       of the parameter "pAppendFlag" (True or False). It returns an integer representing the success of the operation.    

    Parameter(s):
       pOutputFile - string containing the name of the output file where the data will be saved.
       pDataList - list containing the data to be written in the output file.
       pAppendFlag - boolean that specifies if the file is to be appended (True) or overwritten (False) with the new data. If not
                     provided, this parameter is set to False.
    
    Return(s):
       int - 0 (the file was successfully written) or -1 (error found when writing the file).      
    
    Example:
       (...)
       dataList = []
       dataList.append("This is the first line!")
       dataList.append("This is the second line!")
       dataList.append("This is the third line!")
       dataList.append("Enough of lines!")
       
       code = saveOutput("myOutput.txt", dataList)   # write data stored in dataList in the file "myOutput.txt"
       
       if code == 0:
          print "The data was successfully written in the output file."
       else:
          print "Error found when writing the data in the output file."
       (...)    
    """
    
    try:
        #check if path exist if not - create it
        dirPath = os.path.dirname(pOutputFile)
        if dirPath != "" and not os.path.exists(dirPath):
            os.makedirs(dirPath)
            
        if pAppendFlag:                                           
            fileHandler = open(pOutputFile, "a")
        else:
            fileHandler = open(pOutputFile, "w")
        for data in pDataList:
            fileHandler.write("%s\n" % data)
        fileHandler.close()
        return 0
    except:
        return -1     
    


def _applyReplace(pText, pReplace, pMultipleReplace, pReferenceList, pDataList):
    if pReplace is None:
        return pText, []
    else:
        messageList = []
        replaceList0 = []
        lastPosition = 0
        result = ""
        
        while(True):
            i = pReplace.find(",", lastPosition)
            if i == -1:
                replaceList0.append("%s%s" % (result, pReplace[lastPosition:]))
                break
            else:
                if i == 0 or pReplace[i - 1] != "\\":
                    replaceList0.append("%s%s" % (result, pReplace[lastPosition:i]))
                    result = ""
                else:
                    result = "%s%s," % (result, pReplace[lastPosition:i - 1])
                lastPosition = i + 1
        
        singleReplaceFlag = pMultipleReplace is None or pMultipleReplace.upper() != "TRUE"       
        for replace in replaceList0:
            #To handle colon with escape character
            replace = replace.replace("\:","&EscapedColon&")
            replaceList1 = replace.split(":", 3)
            for i in range(len(replaceList1)):
                replaceList1[i] = replaceList1[i].replace("&EscapedColon&",":")

            if len(replaceList1) == 1:
                if pReferenceList is not None:
                    replaceList1[0] = _replaceReference(replaceList1[0], pReferenceList, pDataList, True)
                pText, tmpMessageList = _applyCompute(replaceList1[0])
                messageList.extend(tmpMessageList)
                if singleReplaceFlag: 
                    break                
            else:
                if pReferenceList is not None:
                    replaceList1[0] = _replaceReference(replaceList1[0], pReferenceList, pDataList, True)
                    replaceList1[1] = _replaceReference(replaceList1[1], pReferenceList, pDataList, False)
                if replaceList1[0].lstrip().startswith("%compute(") and replaceList1[0].rstrip().endswith(")%"):                 
                    i = replaceList1[0].find("%compute(") + 9
                    j = replaceList1[0].rfind(")%")             
                    try:
                        if eval(replaceList1[0][i:j]):
                            pText = replaceList1[1]
                            if singleReplaceFlag:
                                break
                    except:
                        messageList.append("Error when computing expression '%s'." % replaceList1[0][i:j])
                else:
                    try:
                        matchList = re.search(replaceList1[0], pText)                    
                        if matchList is not None:
                            lastPosition = 0
                            result = ""
                            while(True):
                                i = replaceList1[1].find("$", lastPosition)
                                if i == -1:
                                    result = "%s%s" % (result, replaceList1[1][lastPosition:])
                                    break
                                else:
                                    j = replaceList1[1].find("$", i + 1)
                                    if j == -1:
                                        result = "%s%s" % (result, replaceList1[1][lastPosition:])
                                        break
                                    else:
                                        try:
                                            result = "%s%s%s" % (result, replaceList1[1][lastPosition:i], matchList.group(int(replaceList1[1][i + 1:j])))
                                        except:
                                            result = "%s%s" % (result, replaceList1[1][lastPosition:j + 1])
                                            lastPosition = j + 1
                                        lastPosition = j + 1                            
                            result, tmpMessageList = _applyCompute(result)
                            messageList.extend(tmpMessageList)  
                            if len(replaceList1) == 2:
                                regularExpression = re.compile(replaceList1[0])
                                pText, count = regularExpression.subn(result, pText)
                            else:
                                try:                           
                                    i = int(replaceList1[2])
                                    if i == -1:
                                        pText = result
                                        count = 1
                                    else:
                                        regularExpression = re.compile(replaceList1[0])
                                        pText, count = regularExpression.subn(result, pText, i)                            
                                except:
                                    count = 0
                            if count > 0 and singleReplaceFlag:
                                break
                    except:
                        messageList.append("Error when computing regular expression '%s'." % replaceList1[0])

        return pText, messageList    

    

def _applySubString(pText, pSubString):
    
    messageList = []
    if pSubString is not None:
        subStringList = pSubString.split(":", 2)
        try:
            if len(subStringList) == 1:
                pText = pText[int(subStringList[0].strip())]
            else:
                if subStringList[0] == "":
                    start = 0                 
                else:
                    start = int(subStringList[0].strip())
                if subStringList[1] == "":
                    end = len(pText)                 
                else:
                    end = int(subStringList[1].strip())
                pText = pText[start:end] 
        except:
            messageList.append("Error when applying substring '%s'." % pSubString)                    
                                                                                   
    return pText, messageList



def _applyUnique(pTextList, pUnique):
    
    messageList = []
    if pUnique is not None:
        if pUnique.upper() == "TRUE":
            uniqueList = []
            for text in pTextList:
                if text not in uniqueList:
                    uniqueList.append(text)
            pTextList = uniqueList
        else:
            uniqueList = pUnique.split(":", 2)
            try:
                if len(uniqueList) == 1:
                    start = int(uniqueList[0].strip())
                    flag0 = 0
                else:
                    if uniqueList[0] == "":
                        start = 0                 
                    else:
                        start = int(uniqueList[0].strip())                        
                    if uniqueList[1] == "":
                        flag0 = 1
                    else:
                        end = int(uniqueList[1].strip())
                        flag0 = 2            
                uniqueList = []
                for text in pTextList:
                    if flag0 == 0:
                        tmpText = text[start]
                    else:
                        if flag0 == 1:
                            tmpText = text[start:]
                        else:
                            tmpText = text[start:end]
                    flag1 = True                    
                    for unique in uniqueList:
                        if flag0 == 0:
                            if tmpText == unique[start]:
                                flag1 = False
                                break                 
                        else:
                            if flag0 == 1:
                                if tmpText == unique[start:]:
                                    flag1 = False
                                    break                 
                            else:
                                if tmpText == unique[start:end]:
                                    flag1 = False
                                    break
                    if flag1:
                        uniqueList.append(text)
                pTextList = uniqueList
            except:
                messageList.append("Error when applying unique '%s'." % pUnique)                    
                                                                                   
    return pTextList, messageList



def _applyCompute(pExpression):
    if pExpression is None:
        return None, []
    else:
        messageList = []
        lastPosition = 0
        result = ""
        while(True):
            i = pExpression.find("%compute(", lastPosition)
            if i == -1:
                result = "%s%s" % (result, pExpression[lastPosition:])
                break
            else:
                j = pExpression.find(")%", i + 9)
                if j == -1:
                    result = "%s%s" % (result, pExpression[lastPosition:])
                    break
                else:
                    result = "%s%s" % (result, pExpression[lastPosition:i])
                    expression = pExpression[i + 9:j]
                    try:
                        exec("flex_compute_result = %s" % expression)
                        result = "%s%s" % (result, flex_compute_result)
                    except:
                        messageList.append("Error when computing expression '%s'." % expression)
                    lastPosition = j + 2
        return result, messageList



def _applyRange(pDataList, pRange):    
    
    if pRange is None:
        return pDataList, []
    else:
        rangeList = pRange.replace("-", ":").split(",")
        messageList = []
        result = []        
        for item in rangeList:
            item = item.strip()
            sliceList = item.split(":")
            i = len(sliceList)
            try:
                if i == 1:                                  
                    result.append(pDataList[int(item) - 1])
                else:
                    start = sliceList[0].strip()
                    if start == "":
                        start = 0
                    else:
                        start = int(start) - 1   
                    end = sliceList[1].strip()
                    if end == "":
                        end = len(pDataList)
                    else:
                        end = int(end)                        
                    if i == 2:
                        step = 1
                    else:
                        step = int(sliceList[2].strip())
                    result.extend(pDataList[start:end:step])
            except:
                messageList.append("Error when applying range '%s'." % pRange)
        return result, messageList  
    



def _displayMessage(pMessage, pForceFlag = False):
    if pMessage.startswith("DEBUG_MODE_"):
        if not globals()["_DEBUG_MODE"]:
            return
        else:
            pMessage = pMessage[11:] 
    if globals()["_UNIQUE_FLAG"]:
        flag = (pMessage != globals()["_LOG_MESSAGE"])
    else:
        flag = True
    if flag:
        now = datetime.now()
        message = "[%02d:%02d:%02d] %s" % (now.hour, now.minute, now.second, pMessage)        
        if not globals()["_QUIET_FLAG"] or pForceFlag:
            print message
            logHandler = globals()["_LOG_HANDLER"]
            if logHandler:
                try:
                    logHandler.write("%s\n" % message)
                    logHandler.flush()
                except:
                    pass             
        globals()["_LOG_MESSAGE"] = pMessage
        



def _showHeader():
    i = len(__copyright__)
    print
    print "*" * (i + 6) 
    print "*  %s  *" % (" " * i)
    print "*  FlexExtractor v%s (%s - %s)%s  *" % (__version__, __date__[7:26], __status__, " " * (i - len(__version__) - len(__date__[7:26]) - len(__status__) - 21))
    print "*  %s  *" % (" " * i)
    print "*  %s%s  *" % (__email__, " " * (i - len(__email__)))
    print "*  %s  *" % (" " * i)
    print "*  %s  *" % __copyright__
    print "*  %s  *" % (" " * i)
    print "*" * (i + 6)            
    print



def errorLogger():
    fp = StringIO.StringIO()
    traceback.print_exc(file=fp)
    return "DEBUG_MODE_" + fp.getvalue()


 

# =============================================
#  MAIN FUNCTION
# =============================================
def _executeMain():

       
    # =============================================
    #  LOCAL VARIABLES
    # =============================================
    showAutoSelectFlag = False
    preProcessFlag = False
    postProcessFlag = False
    configFileFlag = False
    inputFileFlag = False
    headFileFlag = False
    tailFileFlag = False  
    appendFlag = None
    inputType = None
    inputParameter = None
    inputRange = None
    outputFile = None        
    outputType = None
    outputParameter = None    
    preProcess = ""
    postProcess = ""
    configFile = ""
    inputFile = ""
    parameter = ""    
    logFile = ""    
    headFile = ""
    tailFile = ""
    separator = ','
    

    
    # =============================================
    #  CHECK IF PARAMETERS ARE VALID
    # =============================================    
    for argument in sys.argv[1:]:
        if argument.upper() == "-HELP":
            _showHeader()
            print "  -help                Show this help."
            print "  -quiet               Only warning/error messages will be displayed."           
            print "  -unique              Display only unique messages."
            print "  -showautoselect      Display source header if autoselect is enabled."           
            print "  -logfile:x           Append all messages to log file 'x'."           
            print "  -preprocess:x        Execute 'x' before the processing starts."
            print "  -postprocess:x       Execute 'x' after the processing is done."
            print "  -config:x            Use content 'x' as the XML config."
            print "  -configfile:x        Use file 'x' as the XML config."
            print "  -inputfile:x         Read input from file 'x'."
            print "  -inputtype:x         Read input using plugin 'x'."
            print "  -inputparameter:x    Pass parameter 'x' to plugin for reading input."
            print "  -inputrange:x        Define input data range 'x'."
            print "  -head:x              Insert content 'x' at the start of the output."
            print "  -headfile:x          Insert content of file 'x' at the start of the output."                                                                                  
            print "  -tail:x              Insert content 'x' at the end of the output."
            print "  -tailfile:x          Insert content of file 'x' at the end of the output."                                                                                   
            print "  -outputfile:x        Save output to file 'x'."            
            print "  -appendfile:x        Append output to file 'x'."
            print "  -outputtype:x        Write output using plugin 'x'."
            print "  -outputparameter:x   Pass parameter 'x' to plugin for writing output."
            print "  -parameter:x         Assign value 'x' to special keywords '%parameter.X%', where X is number of parameter."      
            print "  -paramseparator:x    Use 'x' to split multiple values in 'parameter'. Default: ','"  
            print "  -debug               Display detailed errors"                
            print
            return 0
        elif argument.upper() == "-QUIET":
            globals()["_QUIET_FLAG"] = True
        elif argument.upper() == "-UNIQUE":
            globals()["_UNIQUE_FLAG"] = True
        elif argument.upper() == "-SHOWAUTOSELECT":
            globals()["_SHOWAUTOSELECT_FLAG"] = True                        
        elif argument[:9].upper() == "-LOGFILE:":
            globals()["_LOG_FLAG"] = True
            logFile = argument[9:]            
        elif argument[:12].upper() == "-PREPROCESS:":
            preProcessFlag = True
            preProcess = argument[12:]
        elif argument[:13].upper() == "-POSTPROCESS:":
            postProcessFlag = True
            postProcess = argument[13:]
        elif argument[:8].upper() == "-CONFIG:":
            if not configFileFlag:
                configFile = argument[8:]
        elif argument[:12].upper() == "-CONFIGFILE:":
            configFile = argument[12:]
            configFileFlag = True
        elif argument[:11].upper() == "-INPUTFILE:":
            inputFile = argument[11:]
            inputFileFlag = True
        elif argument[:11].upper() == "-INPUTTYPE:":           
            inputType = argument[11:]
        elif argument[:16].upper() == "-INPUTPARAMETER:":           
            inputParameter = argument[16:]
        elif argument[:12].upper() == "-INPUTRANGE:":           
            inputRange = argument[12:]            
        elif argument[:6].upper() == "-HEAD:":
            if not headFileFlag:
                headFile = argument[6:].decode("string_escape") #to handle new line
        elif argument[:10].upper() == "-HEADFILE:":
            headFile = argument[10:]
            headFileFlag = True            
        elif argument[:6].upper() == "-TAIL:":
            if not tailFileFlag:
                tailFile = argument[6:]
        elif argument[:10].upper() == "-TAILFILE:":
            tailFile = argument[10:]
            tailFileFlag = True                                   
        elif argument[:12].upper() == "-OUTPUTFILE:":
            if appendFlag is None or not appendFlag:
                outputFile = argument[12:]           
        elif argument[:12].upper() == "-APPENDFILE:":
            outputFile = argument[12:]
            appendFlag = True
        elif argument[:12].upper() == "-OUTPUTTYPE:":           
            outputType = argument[12:]
        elif argument[:17].upper() == "-OUTPUTPARAMETER:":           
            outputParameter = argument[17:]
        elif argument[:11].upper() == "-PARAMETER:":           
            parameter = argument[11:]
        elif argument[:16].upper() == "-PARAMSEPARATOR:":
            separator = argument[16:]
        elif argument.upper() == "-DEBUG":
            globals()["_DEBUG_MODE"] = True
        else:
            _showHeader()
            print "Parameter '%s' invalid. Please execute with '-help' to see valid parameters." % argument
            return -1 
            
    # =============================================
    #  SHOW HEADER (IF PARAMETER "quiet" WAS NOT SPECIFIED)
    # =============================================
    if not _QUIET_FLAG:
        _showHeader()
        
        
                    
    # =============================================
    #  SAVE LOG MESSAGES IF SPECIFIED
    # =============================================        
    if globals()["_LOG_FLAG"]:
        if logFile == "":
            print "Please specify the log file (example: -logfile:/users/dummy/myLog.txt)."
            return -1
        else:
            try:
                globals()["_LOG_HANDLER"] = open(logFile, "a")
            except:
                globals()["_LOG_HANDLER"] = None
                
                
                
    # =============================================
    #  CHECK IF PREPROCESS PARAMETER IS SPECIFIED
    # =============================================        
    if preProcessFlag:
        if preProcess == "":
            print "Please specify the command to be executed (example: -preprocess:dir)."
            return -1



    # =============================================
    #  CHECK IF POSTPROCESS PARAMETER IS SPECIFIED
    # =============================================        
    if postProcessFlag:
        if postProcess == "":
            print "Please specify the command to be executed (example: -postprocess:dir)."
            return -1
                


    # =============================================
    #  CHECK IF CONFIG FILE IS SPECIFIED
    # =============================================
    if configFileFlag:
        if configFile == "":
            print "Please specify the config file (example: -configfile:/users/dummy/myConfig.xml)."
            return -1
        if not os.path.isfile(configFile):
            print "The config file '%s' does not exists." % configFile
            return -1
    else:
        if configFile == "":
            print "Please specify the config (example: -config:\"<flexextractor> <input>...</input> <output>...</output> </flexextractor>\")."
            return -1
        

        
    # =============================================
    #  CHECK IF INPUTRANGE PARAMETER IS SPECIFIED
    # =============================================      
    if inputRange is not None and inputRange == "":
        print "Please specify the input range to be used (example: -inputrange:5:25)."
        return -1
        

    
    try:    
        # =============================================
        #  READ XML CONFIG FILE
        # =============================================        
        _displayMessage("Start processing.")
        if configFileFlag:
            _displayMessage("Reading config file '%s'." % configFile)        
            status, headerList, inputList, outputList = getConfig(configFile)       
            if status == -1:
                _displayMessage("Error when reading config file '%s'." % configFile, True)
                return -1
            elif status == -2:    
                _displayMessage("Error when parsing config file '%s' in line '%s' at column '%s'." % (configFile, headerList, inputList), True)
                return -1
            elif status == -3:
                _displayMessage("Config file '%s' doesn't contain source section." % configFile, True)
                return -1
        else:
            status, headerList, inputList, outputList = getConfigFromText(configFile)    
            if status == -1:
                _displayMessage("Error when reading config '%s'." % configFile, True)
                return -1
            elif status == -2:    
                _displayMessage("Error when parsing config '%s' in line '%s' at column '%s'." % (configFile, headerList, inputList), True)
                return -1      
            
        if not preProcessFlag:
            if headerList["preprocess"] is not None:
                preProcessFlag = True
                preProcess = headerList["preprocess"]
                
        if not postProcessFlag:
            if headerList["postprocess"] is not None:
                postProcessFlag = True
                postProcess = headerList["postprocess"]            
                
    
        # =============================================
        #  EXECUTE PREPROCESS COMMAND IF SPECIFIED
        # =============================================        
        if preProcessFlag:
            _displayMessage("Executing '%s'." % preProcess)
            try:
                subprocess.call(preProcess, shell = True)
            except:
                _displayMessage("Error when executing '%s'." % preProcess, True)
        
        
        # =============================================
        # PREPARE PARAMETER LIST       
        # =============================================
        tmpParamList = parameter.replace('\\'+separator,"&EscapedParamSeparator&").split(separator)
        for i in range(len(tmpParamList)):
            tmpParamList[i] = tmpParamList[i].replace("&EscapedParamSeparator&",separator)
               

        # =============================================
        #  DEFINE KEYWORD LIST
        # =============================================
        keywordList = {}
        if configFileFlag:
            keywordList["configfile"] = [configFile, _STRING]
        else:
            keywordList["configfile"] = ["", _STRING]        
        keywordList["inputfile"] = [inputFile, _STRING]
        if outputFile is None:
            if outputList["file"] is None:
                keywordList["outputfile"] = ["", _STRING]
            else:
                keywordList["outputfile"] = [outputList["file"], _STRING]
        else:
            keywordList["outputfile"] = [outputFile, _STRING]
        if len(tmpParamList) == 1:
            keywordList["parameter"] = keywordList["parameter.0"] = [parameter, _STRING] #raw parameter entered by user
            keywordList["parameter.1"] = [tmpParamList[0], _STRING]
        elif len(tmpParamList) > 1:
            i = 1
            for param in tmpParamList:
                keywordList["parameter.%s" % i] = [param, _STRING]
                i += 1 
            keywordList["parameter"] = keywordList["parameter.0"] = [parameter, _STRING] #raw parameter entered by user
    
    
        # =============================================
        #  READ DATA FROM INPUT FILE(S) USING PLUGIN(S) 
        # =============================================
        if inputFileFlag:
            if inputFile == "":
                _displayMessage("Please specify the input file (example: -inputfile:/users/dummy/myInput.xls).", True)
                return -1            
            else:
                status, dataList, messageList = getData(inputList, keywordList, inputFile, inputType, inputParameter, inputRange)
        else:           
            status, dataList, messageList = getData(inputList, keywordList, None, inputType, inputParameter, inputRange)
        for message in messageList:
            _displayMessage(message, True)    
        if status == -1:
            return -1       
    
        # =============================================
        #  GENERATE AND SAVE/DISPLAY OUTPUT
        # =============================================
        _displayMessage("Generating output.")        
            
        if headFile == "":
            headList = None
        else:
            if headFileFlag:
                try:
                    fileHandler = open(headFile, "r")
                    headList = fileHandler.read().split("\n");
                    fileHandler.close()                           
                except:
                    headList = []
                    _displayMessage("Error when reading head file '%s'." % headFile, True)
            else:
                headList = [headFile]
        
        if tailFile == "":
            tailList = None
        else:
            if tailFileFlag:
                try:
                    fileHandler = open(tailFile, "r")
                    tailList = fileHandler.read().split("\n");            
                    fileHandler.close()                           
                except:
                    tailList = []
                    _displayMessage("Error when reading tail file '%s'." % tailFile, True)
            else:
                tailList = [tailFile]              
        
        if outputFile is not None and appendFlag is None:
            appendFlag = False            
        
#        keywordList = {}
#        if configFileFlag:
#            keywordList["configfile"] = [configFile, _STRING]
#        else:
#            keywordList["configfile"] = ["", _STRING]        
#        keywordList["inputfile"] = [inputFile, _STRING]
#        if outputFile is None:
#            if outputList["file"] is None:
#                keywordList["outputfile"] = ["", _STRING]
#            else:
#                keywordList["outputfile"] = [outputList["file"], _STRING]
#        else:
#            keywordList["outputfile"] = [outputFile, _STRING]
#        keywordList["parameter"] = [parameter, _STRING]            
            
        code, resultList, messageList = generateOutput(inputList, outputList, dataList, keywordList, outputFile, outputType, outputParameter, appendFlag, headList, tailList)        
        for message in messageList:
            _displayMessage(message, True)
        
#        if outputFile is None:
#            if outputList["file"] is None:
#                outputFile = ""
#            else:
#                outputFile = outputList["file"]            
#        
#        if outputFile == "":
#            for result in resultList:
#                print result                
#        else:
#            if outputType is None and outputList["type"] is None:
#                if appendFlag:
#                    _displayMessage("Appending output to file '%s'." % outputFile)
#                else:
#                    _displayMessage("Writing output to file '%s'." % outputFile)   
#                if saveOutput(outputFile, resultList, appendFlag) == -1:
#                    _displayMessage("Error when writing output to file '%s'." % outputFile, True)
#                    return -1
                
    
    
        # =============================================
        #  EXECUTE POSTPROCESS COMMAND IF SPECIFIED
        # =============================================        
        if postProcessFlag:
            _displayMessage("Executing '%s'." % postProcess)
            try:
                subprocess.call(postProcess, shell = True)
            except:
                _displayMessage("Error when executing '%s'." % postProcess, True)    
            
        
    
        # =============================================
        #  TIME TO SAY GOODBYE AND EXIT GRACEFULLY! :)
        # =============================================
        _displayMessage("Processing done.")
        
        returnCode = 0        
     
        
    except KeyboardInterrupt:
        _displayMessage("Processing interrupted by the user.", True)
        returnCode = 0
    except:
        _displayMessage("An unexpected error was found while processing.", True)
        returnCode = -1
           
    if globals()["_LOG_HANDLER"] is not None:
        globals()["_LOG_HANDLER"].close()
                         

    return returnCode



 
if __name__ == "__main__":


    # =============================================
    #  PROFILING (CREATES A FILE CALLED "FlexExtractor.prof"
    #  CONTAINING STATISTICS ABOUT FlexExtractor EXECUTION).
    #
    #  TO ANALYSE THE GENERATED PROFILE FILE, ONE CAN USE
    #  TOOLS LIKE RunSnakeRun (http://www.vrplumber.com/programming/runsnakerun/)
    #  OR Gprof2Dot (http://code.google.com/p/jrfonseca/wiki/Gprof2Dot).
    #
    #  TO EXECUTE RunSnakeRun FROM A TERMINAL:
    #
    #     python runsnake.py FlexExtractor.prof
    # 
    #  TO EXECUTE Gprof2Dot FROM A TERMINAL:
    #
    #     python gprof2dot.py -f pstats FlexExtractor.prof | dot -Tpng -o FlexExtractor.png  
    # =============================================
    #import cProfile
    #command = """_executeMain()"""
    #cProfile.runctx(command, globals(), locals(), filename = "FlexExtractor.prof")
    #sys.exit(0)

    # command = """_executeMain()"""
    # import profile
    # profile.run(command, sort='time')
    # sys.exit(0)

    # import timeit
    # print timeit.Timer("from __main__ import _executeMain;"+command).timeit(number=1)

    # =============================================
    #  EXECUTE MAIN FUNCTION
    # =============================================   
    sys.exit(_executeMain())
    

