#!/usr/bin/env python

__author__ = "Bartlomiej Tomasz Urbaniec"
__email__ = "bartlomiej.tomasz.urbaniec@cern.ch"
__copyright__ = "(C) 2011 European Organization for Nuclear Research, Geneva, Switzerland"
__version__ = "2.5"
__date__ = "$Date: 2012-03-19 16:47:14 +0100 (Mon, 19 Mar 2012) $"
__description__ = "Plugin for FlexExtractor to handle SPEC XML files"
__status__ = "Production" 
 

def _returnValue(node):
    if len(node) > 0:
        return node[0]
    else:
        return None

# =============================================
#  FUNCTION RESPONSIBLE FOR READING DATA FROM A CERTAIN
#  INPUT. THIS FUNCTION IS CALLED BY FlexExtractor.
# =============================================
def getData(pFile, pSourceList):
    """
    This function reads data from an input source specified in the parameter "pFile". The structure
    defined in "pSourceList" is used to know what data to read from the input source. This function
    returns a tuple with four elements: code, header, data and message.

    Parameter(s):
       pFile - string containing the name of the input source to be read.     
       pSourceList - list containing the input structure that was specified in the XML config.
    
    Return(s):
       int - 0 (the input source was successfully read) or -1 (error found when reading the input source).
       list - contains a list with the headers of the input source (this list should be empty when the "autoselect"
              is False and filled-up when "autoselect" is True).                    
       list - contains the data that was read from the input source (this list will be empty if some error was found).
       list - contains a list of messages that were generated during the reading of the input source.             
    """
    try:
        import FlexExtractor
    except:
        return -1, ["Error when importing module 'FlexExtractor'."]  
        
    try:
        import os
    except:
        return -1, [], [], ["Error when importing module 'os'."]

    try:
        import re
    except:
        return -1, [], [], ["Error when importing module 're'."]

    if not os.path.isfile(pFile):
        return -1, [], [], ["The input file '%s' does not exists." % pFile]    
    
    try:
        from lxml import etree
    except:
        return -1, [], [], ["Error when importing module 'etree'."]              

      
    try:
        nameSpaces = {'ss': 'urn:schemas-microsoft-com:office:spreadsheet'}
        NS = '{urn:schemas-microsoft-com:office:spreadsheet}'
        copyStyles = False
        parserOptions = etree.XMLParser(strip_cdata=False)
        parser = etree.parse(pFile,parserOptions)  
        autoSelectList = []
        messageList = []        
        result = []
        for source in pSourceList:
            # =============================================
            # GET DATA START IF RANGE IS NONE
            # =============================================
            if source["range"] is None:
                tmpHE = _returnValue(parser.xpath("/ss:Workbook/ss:Worksheet[@ss:Name='%s']/ss:Table/@ss:HeaderEnd" % source["name"], namespaces=nameSpaces))
                if tmpHE is not None:
                    tmpHE = int(tmpHE) + 1
                    source["range"] = str(tmpHE) + ":"
            
            # =============================================
            # COPY ALL STYLES FROM INPUT TO OUTPUT
            # =============================================
            if source["name"].upper() == "COPYSTYLES":
                copyStyles = True
                stylesList = parser.xpath("/ss:Workbook/ss:Styles/ss:Style", namespaces=nameSpaces)
                tmpResult = []
                tmpResult.append(["COPYSTYLES"])
                for style in stylesList:
                    tmpResult.append([etree.tostring(style)])
                result.append(tmpResult)
                autoSelectList.append(["data"]) 
                continue                           
            
            # =============================================
            # COPY ALL SHEETS THAT HAVEN'T BEEN SPECIFIED 
            # IN CONFIGURATION
            # =============================================
            if source["name"].upper() == "COPYALLSHEETS": 
                # Distinguishing between sheets which exist and don't exist in configuration file  
                sheetNames = []
                for sh in pSourceList:
                    if sh["name"].upper() == "COPYALLSHEETS" or sh["name"].upper() == "COPYSTYLES":
                        continue
                    sheetNames.append(sh["name"])
                    
                sheetListInput = parser.xpath("/ss:Workbook/ss:Worksheet", namespaces=nameSpaces)
                tmpSheetListInput = sheetListInput[:]
                for sheetInput in tmpSheetListInput:
                    tmpName =  _returnValue(sheetInput.xpath(".//@ss:Name", namespaces=nameSpaces))
                    if tmpName in sheetNames:
                        sheetListInput.remove(sheetInput)
                        
                # Preparing sheets and rows which will be copy to the output
                tmpResult = []
                for sheetInput in sheetListInput:
                    tmpName =  _returnValue(sheetInput.xpath(".//@ss:Name", namespaces=nameSpaces))
                    tmpResult.append(["SHEETMOVE:"+tmpName])
                    rowListInput = sheetInput.xpath("ss:Table/ss:Row", namespaces=nameSpaces)
                    for row in rowListInput:
                        if NS+"Hidden" in row.attrib:
                            del row.attrib[NS+"Hidden"]
                        if not copyStyles:
                            if NS+"StyleID" in row.attrib:
                                del row.attrib[NS+'StyleID']
                        cellList = row.xpath("ss:Cell", namespaces=nameSpaces)    
                        if not copyStyles:
                            for cell in cellList:
                                if NS+"StyleID" in cell.attrib:
                                    del cell.attrib[NS+'StyleID']
                        tmpRow = etree.tostring(row)
                        tmpRow = tmpRow.replace('xmlns="urn:schemas-microsoft-com:office:spreadsheet"','')
                        tmpResult.append([tmpRow])
                result.append(tmpResult)
                autoSelectList.append(["data"]) 
                continue
                       
            rowList = parser.xpath("/ss:Workbook/ss:Worksheet[@ss:Name='%s']/ss:Table/ss:Row" % source["name"], namespaces=nameSpaces)
            if rowList == []:
                autoSelectList.append([])                
                result.append([])
                messageList.append("The input file '%s' does not have a sheet '%s' or it is empty." % (pFile, source["name"]))
            

            elif source["parameter"] is not None and source["parameter"].upper() == "NOTDEVICETYPE":
                data = source["data"]
                if len(data) == 0:
                    result.append([])
                    autoSelectList.append([])
                else:
                    selectedColumns = []
                    for d in data:
                        selectedColumns.append(d["select"])
                    tmpResult = []
                    for row in rowList:
                        tmpRow = [""] * len(selectedColumns)
                        cellList = row.xpath(u"./ss:Cell", namespaces=nameSpaces)
                        j = 0
                        for column in selectedColumns:
                            i = 1
                            for cell in cellList:
                                index = _returnValue(cell.xpath(".//@ss:Index", namespaces=nameSpaces))
                                if index is not None:
                                    i = int(index)
                                if i == int(column):
                                    value = _returnValue(cell.xpath("./ss:Data/text()", namespaces=nameSpaces))
                                    if value is not None:
                                        tmpRow[j] = value
                                if i > column:
                                    break
                                i += 1
                            j += 1
                        tmpResult.append(tmpRow)
                    autoSelectList.append([])
                    result.append(tmpResult)
            else:
                sourceAutoSelect = source["autoselect"]
                flag = True
                if sourceAutoSelect is None:
                    autoSelectFlag = False           
                    startHeader = _returnValue(parser.xpath("/ss:Workbook/ss:Worksheet[@ss:Name='%s']/ss:Table/@ss:HeaderStart" % source["name"], namespaces=nameSpaces))
                    endHeader = _returnValue(parser.xpath("/ss:Workbook/ss:Worksheet[@ss:Name='%s']/ss:Table/@ss:HeaderEnd" % source["name"], namespaces=nameSpaces))
                    if startHeader is None or endHeader is None:
                        autoSelectList.append([])
                        result.append([])
                        messageList.append("The sheet '%s' doesn't have the 'HeaderStart' or 'HeaderEnd' attributes defined." % source["name"])
                        continue                        
                        #return -1, [], [], ["The sheet '%s' doesn't have the 'HeaderStart' or 'HeaderEnd' attributes defined." % source["name"]]
                    else:                        
                        startHeader = int(startHeader) - 1
                        endHeader = int(endHeader)                                                          
                else:
                    autoSelectFlag = True

                    if sourceAutoSelect.upper() == "TRUE":                                  
                        startHeader = _returnValue(parser.xpath("/ss:Workbook/ss:Worksheet[@ss:Name='%s']/ss:Table/@ss:HeaderStart" % source["name"], namespaces=nameSpaces))
                        endHeader = _returnValue(parser.xpath("/ss:Workbook/ss:Worksheet[@ss:Name='%s']/ss:Table/@ss:HeaderEnd" % source["name"], namespaces=nameSpaces))
                        if startHeader is None or endHeader is None:
                            autoSelectList.append([])
                            result.append([])
                            messageList.append("The sheet '%s' doesn't have the 'HeaderStart' or 'HeaderEnd' attributes defined." % source["name"])
                            continue
                            #return -1, [], [], ["The sheet '%s' doesn't have the 'HeaderStart' or 'HeaderEnd' attributes defined." % source["name"]]
                        else:                        
                            startHeader = int(startHeader) - 1
                            endHeader = int(endHeader)                        
                    else:
                        try:           
                            sourceAutoSelectList = sourceAutoSelect.split(":", 2)
                            if len(sourceAutoSelectList) == 1:
                                startHeader = int(sourceAutoSelectList[0]) - 1
                                endHeader = int(sourceAutoSelectList[0])
                            else:
                                startHeader = int(sourceAutoSelectList[0]) - 1
                                endHeader = int(sourceAutoSelectList[1])                                                        
                        except:
                            flag = False
                            autoSelectList.append([])
                            messageList.append("The sheet '%s' has an invalid autoselect '%s' (it should specify the start/end of the header rows)." % (source["name"], sourceAutoSelect))            
                
                tmpData0 = []
                if flag:
                    # =============================================
                    #  FILLING EMPTY ROW IF INDEX EXIST FOR ROW
                    # ============================================= 
                    newRowList = []
                    rowIterator = 1
                    for tmpRow in rowList:
                        rowIndex = _returnValue(tmpRow.xpath("@ss:Index", namespaces=nameSpaces))
                        if rowIndex is not None and rowIterator <= int(rowIndex):
                            while rowIterator < int(rowIndex):  
                                newRowList.append(etree.Element("Row",nsmap=nameSpaces))
                                rowIterator += 1
                        newRowList.append(tmpRow)
                        rowIterator += 1
                    rowList = newRowList
                                        
                    # =============================================
                    #  GET HEADER
                    # =============================================                    
                    tmpHeaderList0 = []
                    for row in rowList[startHeader:endHeader]:
                        cellList = row.xpath(u"./ss:Cell", namespaces=nameSpaces)
                        tmpHeaderList1 = []               
                        for cell in cellList:
                            value = _returnValue(cell.xpath("./ss:Data/text()", namespaces=nameSpaces))
                            if value is None and row==rowList[startHeader]: #skip empty columns in first row of header
                                continue
                            elif value is None:
                                value = ""
                            mergeAcross = _returnValue(cell.xpath(".//@ss:MergeAcross", namespaces=nameSpaces)) 
                            if mergeAcross is None:
                                mergeAcross = 1
                            else:
                                mergeAcross = int(mergeAcross) + 1
                            for i in range(mergeAcross):
                                tmpHeaderList1.append(value)
                        tmpHeaderList0.append(tmpHeaderList1)
                    tmpAutoSelectList = []             
                    headerList = {}                
                    if tmpHeaderList0 != []:
                        for i in range(len(tmpHeaderList0[0])):
                            header = ""
                            for j in range(len(tmpHeaderList0)):
                                if header == "":
                                    header = "%s" % tmpHeaderList0[j][i]
                                else:
                                    if tmpHeaderList0[j][i] != "":
                                        header = "%s.%s" % (header, tmpHeaderList0[j][i])
                            headerList[header] = i
                            for data in source["data"]:
                                if data["regexpselect"] is not None and data["regexpselect"].upper() == "TRUE":
                                    try:
                                        if re.search(data["select"], header) is not None:
                                            data["select"] = header
                                    except:
                                        pass                                    
                            tmpAutoSelectList.append(header)
                    autoSelectList.append(tmpAutoSelectList)
                    

                    # =============================================
                    #  GET DATA
                    # =============================================
                    i = len(headerList)
                    rowIterator = 1
                    for row in rowList:                    
                        rowStyleID = _returnValue(row.xpath("@ss:StyleID", namespaces=nameSpaces))
                        cellList = row.xpath("./ss:Cell", namespaces=nameSpaces)           
                        tmpData1 = [""] * i                  
                        j = 0                    
                        for cell in cellList:
                            cellStyleID = _returnValue(cell.xpath("@ss:StyleID", namespaces=nameSpaces))
                            if cellStyleID is None:
                                if rowStyleID is not None:
                                    cellStyleID = "FLEXROW_" + rowStyleID
                                else:
                                    cellStyleID = ""
                            value = _returnValue(cell.xpath("./ss:Data/text()", namespaces=nameSpaces))
                            if value is None:
                                value = ""
                            index = _returnValue(cell.xpath("@ss:Index", namespaces=nameSpaces))
                            if copyStyles:
                                value = cellStyleID + "&&&" + value
                            if index is None:
                                j += 1
                            else:  
                                j = int(index)
                            if j <= i:                          
                                tmpData1[j - 1] = value                          
                        if autoSelectFlag:
                            tmpData2 = tmpData1
                        else:
                            tmpData2 = []                    
                        for data in source["data"]:
                            if data["select"] is None:
                                tmpData2.append(None)
                            else:                                     
                                try:
                                    tmpData2.append(tmpData1[headerList[data["select"]]])
                                except:
                                    message = "The input file '%s' does not have a select '%s' in sheet '%s' or it is invalid." % (pFile, data["select"], source["name"])
                                    if message not in messageList:
                                        messageList.append(message)                                
                                    tmpData2.append(None)
                        tmpData0.append(tmpData2)                   
                        rowIterator += 1
                result.append(tmpData0)
        return 0, autoSelectList, result, messageList     
    except MemoryError:
        return -1, [], [], ["Error when reading input file '%s' since there is not enough memory available." % pFile, FlexExtractor.errorLogger()]                    
    except:
        return -1, [], [], ["Error when reading/parsing input file '%s'." % pFile, FlexExtractor.errorLogger() ]


    


# =============================================
#  FUNCTION RESPONSIBLE FOR WRITING DATA TO A CERTAIN
#  OUTPUT. THIS FUNCTION IS CALLED BY FlexExtractor.
# =============================================
def setData(pFile, pParameter, pDataList, pAppendFlag):
    """
    This function writes the data "pDataList" into a certain output specified in the parameter "pFile"
    and uses the parameter "pParameter" to provide additional information on how to do it. If "pAppendFlag"
    is True then the data will be appended in the end of the output, otherwise (if False) the previous
    data will be erased before writing the new data. This function returns a tuple with two elements:
    code and message.

    Parameter(s):
       pFile - string containing the name of the output where to write the data.    
       pParameter - string containing the parameter that was specified in the XML config.    
       pDataList - list containing the data to be written.
       pAppendFlag - boolean that specifies if the output is to be overwritten (True) or appended (False) with the new data.
    
    Return(s):
       int - 0 (the data was successfully written) or -1 (error found when writing the data).             
       list - contains a list of messages that were generated during the writing of the data.      
    """   

    try:
        import FlexExtractor
    except:
        return -1, ["Error when importing module 'FlexExtractor'."]  
    
    try:
        import os
    except:
        return -1, ["Error when importing module 'os'."]        
    
    try:
        from lxml import etree
    except:
        return -1, ["Error when importing module 'etree'."]       
 
    try:
        from StringIO import StringIO
    except:
        return -1, [], [], ["Error when importing module 'StringIO'."]   
            
    if not os.path.isfile(pFile):
        return -1, ["The output file '%s' does not exists." % pFile]
    
    if pParameter is None:
        separator = ","
    else:
        separator = pParameter    
       
       
    messageList = []
    copyStyles = False
    try:
        nameSpaces = {'ss': 'urn:schemas-microsoft-com:office:spreadsheet'}
        NS = '{urn:schemas-microsoft-com:office:spreadsheet}'
        
        #preprocess - it's needed because lxml has a problem with namespaces which has the same urn
        preProcess = open(pFile, "r")
        readFile = preProcess.read()
        preProcess.close()
        
        oldString = 'xmlns="urn:schemas-microsoft-com:office:spreadsheet"'
        readFile = readFile.replace(oldString, "")   
                   
        preProcess = open(pFile, "w")
        preProcess.write(readFile)
        preProcess.close()

        parserOptions = etree.XMLParser(strip_cdata=False)
        parser = etree.parse(pFile,parserOptions)
        
        wbWithout = parser.xpath("/Workbook", namespaces=nameSpaces)
        if len(wbWithout) != 0:
            nsPrefix = ""
        else:
            nsPrefix = "ss:"
        
        sheetList = parser.xpath("/%sWorkbook/%sWorksheet" % (nsPrefix, nsPrefix), namespaces=nameSpaces)
              
        if sheetList == []:
            return -1, ["The output file '%s' does not have sheets." % pFile]
        else:
            i = 0
            delListSheet = {}
            while i < len(pDataList):
                flagAdd = False
                flagCopy = False
                tmpData = pDataList[i].upper()
                if tmpData.startswith("SHEET:"):
                    sheetName = pDataList[i][6:]
                    flag0 = False
                    flagAdd = False
                elif tmpData.startswith("SHEETMOVE:"):
                    sheetName = pDataList[i][10:]
                    flagAdd = True
                    flag0 = False
                elif tmpData.startswith("SHEETNOTDEVICE:"):
                    sheetName = pDataList[i][15:]
                    flagCopy = True
                    flag0 = False
                elif tmpData.startswith("DELETEALLEMPTYSHEETS"):
                    for sheetTemp in sheetList:
                        sheetTempName =  _returnValue(sheetTemp.xpath(".//@ss:Name", namespaces=nameSpaces))
                        if sheetTempName not in delListSheet:
                            if sheetTempName != "ProjectDocumentation": #to prevent deleting ProjectDocumentation
                                delListSheet[sheetTempName] = 1
                    i += 1
                    continue
                elif tmpData.startswith("DELETEEMPTYSHEET:"):
                    tmpName = pDataList[i][17:]
                    if tmpName not in delListSheet:
                        delListSheet[tmpName] = 1
                    i += 1
                    continue
                elif tmpData.startswith("COPYSTYLES"):
                    copyStyles = True
                    style = workbook =  parser.xpath("/%sWorkbook/%sStyles" % (nsPrefix,nsPrefix), namespaces=nameSpaces)
                    i += 1
                    while i < len(pDataList):
                        tmpData = pDataList[i].upper()                           
                        if tmpData.startswith("SHEET:") or tmpData.startswith("SHEETMOVE:") or tmpData.startswith("DELETEALLEMPTYSHEETS") or tmpData.startswith("COPYSTYLES") or tmpData.startswith("SHEETNOTDEVICE"):
                            break
                        styleElementTree = etree.parse(StringIO(pDataList[i]))
                        styleElement = styleElementTree.getroot()
                        # check if styles already exist in output file
                        styleID = styleElement.attrib[NS+'ID']
                        oldStyleID = style[0].xpath("%sStyle[@ss:ID='%s']" % (nsPrefix,styleID), namespaces=nameSpaces)
                        if len(oldStyleID) == 0:
                            style[0].append(styleElement)
                        i += 1
                    continue
                
                else:
                    sheetName = ""               
                if sheetName != "":
                    flag1 = False
                    for sheet in sheetList:
                        if sheetName == _returnValue(sheet.xpath(".//@ss:Name", namespaces=nameSpaces)) and not flagAdd and not flagCopy: 
                            flag1 = True                            
                            tableList = sheet.xpath("%sTable" % nsPrefix, namespaces=nameSpaces)
                            if tableList == []:
                                messageList.append("The sheet '%s' in output file '%s' does not have a table." % (sheetName, pFile))
                            else:
                                table = tableList[0]                   
                                if not pAppendFlag:
                                    j = int(_returnValue(table.xpath(".//@ss:HeaderEnd", namespaces=nameSpaces)))
                                    while len(table.xpath("%sRow" % nsPrefix, namespaces=nameSpaces)) > j:                             
                                        table.remove(table.xpath("%sRow" % nsPrefix, namespaces=nameSpaces)[j])
                                for j in range(i + 1, len(pDataList)):
                                    tmpDel = pDataList[j]
                                    if tmpDel.startswith("DELETEEMPTYSHEET:"):
                                        sheetToBeDeleted = tmpDel[17:]
                                        if sheetToBeDeleted not in delListSheet:
                                            delListSheet[sheetToBeDeleted] = 1
                                        i += 1
                                        continue
                                    tmpData = pDataList[j].upper()                           
                                    if tmpData.startswith("SHEET:") or tmpData.startswith("SHEETMOVE:") or tmpData.startswith("DELETEALLEMPTYSHEETS") or tmpData.startswith("COPYSTYLES") or tmpData.startswith("SHEETNOTDEVICE"):
                                        break
                                    else:                              
                                        rowElement = etree.Element("Row",nsmap=nameSpaces)
                                        rowElement.set(NS + "AutoFitHeight", "0")
                                        dataList = pDataList[j].split(separator)
                                        for data in dataList:
                                            cellElement = etree.Element("Cell",nsmap=nameSpaces)
                                            data = data.decode("utf-8")
                                            tmpData = data.split("&&&")
                                            if len(tmpData) >= 2:
                                                styleID = tmpData[0]
                                                if not copyStyles:
                                                    styleID = ""
                                                if styleID != "":
                                                    if styleID.startswith("FLEXROW_"):
                                                        styleID = styleID[8:]
                                                        rowElement.set(NS+"StyleID", styleID)
                                                    else:
                                                        cellElement.set(NS+"StyleID", styleID)
                                                data = tmpData[1]
                                            result = data
                                            if result != "":
                                                dataElement = etree.Element("Data",nsmap=nameSpaces)
                                                dataElement.set(NS + "Type", "String")
                                                dataElement.text = result                              
                                                cellElement.append(dataElement)
                                            rowElement.append(cellElement)
                                        table.append(rowElement)
                                        delListSheet[sheetName] = 0
                                        i += 1                                                                               
                                if NS+"ExpandedRowCount" in table.attrib:
                                    del table.attrib[NS+"ExpandedRowCount"]
                            break                    
                    if flagAdd:
                        flag1 = True
                        workbook =  parser.xpath("/%sWorkbook" % nsPrefix, namespaces=nameSpaces)
                        worksheet = etree.Element("Worksheet")
                        found = workbook[0].xpath("%sWorksheet[@ss:Name='%s']" % (nsPrefix, sheetName), namespaces=nameSpaces)
                        if len(found) > 0:
                            delListSheet[sheetName] = 0
                        else:
                            worksheet.set(NS + "Name", sheetName)
                            table = etree.Element("Table")
                            worksheet.append(table)
                            workbook[0].append(worksheet)
                            k = 0
                            for k in range(i+1, len(pDataList)):
                                tmpData = pDataList[k].upper()  
                                if tmpData.startswith("SHEET:") or tmpData.startswith("SHEETMOVE:") or tmpData.startswith("DELETEALLEMPTYSHEETS") or tmpData.startswith("COPYSTYLES") or tmpData.startswith("SHEETNOTDEVICE") or tmpData.startswith("DELETEEMPTYSHEET:"):
                                    break
                                else:
                                    delListSheet[sheetName] = 0
                                    rowElementTree = etree.parse(StringIO(pDataList[k]))
                                    rowElement = rowElementTree.getroot()
                                    if copyStyles:
                                        table.append(rowElement)
                                    else:
                                        if NS+"StyleID" in rowElement.attrib:
                                            del rowElement.attrib[NS+'StyleID']    
                                        cellList = rowElement.xpath("Cell", namespaces=nameSpaces)  
                                        for cell in cellList:
                                            if NS+"StyleID" in cell.attrib:
                                                del cell.attrib[NS+'StyleID']
                                        table.append(rowElement)     
                                                                                     
                                    i+=1                          
                            messageList.append("Added sheet '%s' to output file '%s'." % (sheetName,pFile))
                    if flagCopy:
                        flag1 = True
                        rowList = parser.xpath("/%sWorkbook/%sWorksheet[@ss:Name='%s']/%sTable/%sRow" % (nsPrefix,nsPrefix,sheetName,nsPrefix,nsPrefix), namespaces=nameSpaces)
                        line = 0
                        for k in range(i+1, len(pDataList)):
                            tmpData = pDataList[k].upper()  
                            if tmpData.startswith("SHEET:") or tmpData.startswith("SHEETMOVE:") or tmpData.startswith("DELETEALLEMPTYSHEETS") or tmpData.startswith("COPYSTYLES") or tmpData.startswith("SHEETNOTDEVICE") or tmpData.startswith("DELETEEMPTYSHEET:"):
                                break
                            else:
                                delListSheet[sheetName] = 0
                                if len(rowList) > line:
                                    cellList = rowList[line].xpath("%sCell" % nsPrefix, namespaces=nameSpaces)
                                    splitedData = pDataList[k].split(pParameter)
                                    it = 0
                                    for tmpD in splitedData:
                                        if tmpD != "":
                                            if len(cellList) > it:
                                                cellData = cellList[it].xpath("%sData" % nsPrefix, namespaces=nameSpaces)
                                                tmpD = tmpD.decode("utf-8")
                                                if len(cellData) != 0:
                                                    cellData[0].text = tmpD
                                                else:
                                                    dataElement = etree.Element("Data",nsmap=nameSpaces)
                                                    dataElement.set(NS + "Type", "String")
                                                    dataElement.text = tmpD
                                                    cellList[it].append(dataElement)    
                                        it += 1                                
                                line += 1

                            
                            
                            
                            
                    if not flag1 and not flag0:
                        messageList.append("The output file '%s' does not have a sheet '%s'." % (pFile, sheetName))
                i += 1
    
            #Time for delete unused sheets from output file
            
            for delSheet in delListSheet:
                if delListSheet[delSheet] == 1:
                    sheet = parser.xpath("/%sWorkbook/%sWorksheet[@ss:Name = '%s']" % (nsPrefix, nsPrefix, delSheet), namespaces=nameSpaces)
                    flagDelete = True
                    if len(sheet) > 0:
                        table = sheet[0].xpath("%sTable" % nsPrefix, namespaces=nameSpaces)
                        if len(table) > 0:
                            # This part checks if output sheet contains data, this is done by compare EndHeader with numbers of line
                            # If number of rows is higher than EndHeader that means that sheet contains data (not only header)
                            # If attribute header doesn't exist plugin checks if there is at least one row, if yes sheet won't be delete
                            HeaderEnd = int(_returnValue(table[0].xpath(".//@ss:HeaderEnd", namespaces=nameSpaces)))
                            if HeaderEnd is None:
                                HeaderEnd = 0
                            if len(table[0].xpath("%sRow" % nsPrefix, namespaces=nameSpaces)) > HeaderEnd:
                                flagDelete = False
                        if flagDelete:
                            sheet[0].getparent().remove(sheet[0])
                            messageList.append("From output file '%s' sheet '%s' has been deleted." % (pFile, delSheet))
            
            
            
            fileHandler = open(pFile, "w")
            fileHandler.write('<?xml version="1.0" encoding="UTF-8"?>')
            fileHandler.write(etree.tostring(parser, encoding="UTF-8", pretty_print=True))
            fileHandler.close()
            
            #postprocess - it's needed because lxml has a problem with namespaces which have the same urn
            postProcess = open(pFile, "r")
            readFile = postProcess.read()
            postProcess.close()
            oldString = '<Workbook'
            newString = '<Workbook xmlns="urn:schemas-microsoft-com:office:spreadsheet"'
            readFile = readFile.replace(oldString, newString)   
            postProcess = open(pFile, "w")
            postProcess.write(readFile)
            postProcess.close()
            
            return 0, messageList
             
    except:
        if pAppendFlag:
            return -1, ["Error when appending output to file '%s'." % pFile, FlexExtractor.errorLogger()]
        else:
            return -1, ["Error when writing output to file '%s'." % pFile, FlexExtractor.errorLogger()]
   
    
