package org.xdv.clx.base;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xdv.clx.builder.ClxBuilderException;
import org.xdv.clx.expr.ClxMacro;
import org.xdv.clx.expr.ClxMacroDefinitions;
import org.xdv.clx.expr.ClxRules;
import org.xdv.clx.extension.ClxExtMacroCall;
import org.xdv.clx.extension.ClxExtMacroCallPostPrep;
import org.xdv.clx.operator.ClxOperatorCall;
import org.xdv.clx.operator.ClxOperatorCallPostPrep;
import org.xdv.clx.operator.ClxOperatorDefFactory;
import org.xdv.xpath.XPathEngine;
import org.xdv.xpath.jaxen.JaxenXPathEngine;

/**
 * Factory to create {@link ClxValidator} from different sources.
 * {@link ClxValidator} should be always created through this factory.
 *
 * @author Tomasz Kokoszka
 */
public class ClxValidatorFactory {
    
    /** XPath engine to both create XPath expressions and execute them during validaton */
    protected final XPathEngine xpathEngine;
    
    /** Registered user's operators */
    protected final Map operators = new HashMap();
    
    /** Rules loader - loads rule set from different kind of sources (file, stream, url, etc.) */
    protected final ClxLoader rulesLoader;
    
    /** Macros loader to load external macros required by rules - 
     * element {@link ClxConstants#TAG_MACROS} */
    protected ClxMacroLoader macrosLoader;
    
    /**
     * Creates factory with default XPath engine ({@link JaxenXPathEngine}) and
     * default macros loader ({@link ClxLoader}).
     */
    public ClxValidatorFactory() {
        this(new JaxenXPathEngine());        
    }        
    
    /**
     * Creates factory with default macros loader ({@link ClxLoader}).
     * @param xpathEngine XPath engine to both create XPath expressions and execute them during 
     *        validation 
     */
    public ClxValidatorFactory(XPathEngine xpathEngine) {        
        this.xpathEngine = xpathEngine;
        this.rulesLoader = new ClxLoader(this.xpathEngine);
        this.rulesLoader.setSkipHeader(true);
        ClxLoader macros = new ClxLoader(this.xpathEngine);
        macros.setSkipHeader(true);
        this.macrosLoader = macros;
    }
    
    /**
     * Creates factory.
     * @param xpathEngine XPath engine to both create XPath expressions and execute them during 
     *        validation
     * @param macrosLoader macros loader to load external macros required by rules
     */
    public ClxValidatorFactory(XPathEngine xpathEngine, ClxMacroLoader macrosLoader) {
        this.xpathEngine = xpathEngine;
        this.rulesLoader = new ClxLoader(this.xpathEngine);
        this.rulesLoader.setSkipHeader(true);        
        this.macrosLoader = macrosLoader;
    }
    
    /** 
     * Returns <code>true</code> if rule header is skipped by this loader 
     * If flag is <code>true</code> rule header will not be loaded, 
     * but simply set to <code>null</code>.
     * @return Value <code>true</code> means that rule header is skipped by this loader
     */
    public boolean isSkipRuleHeader() {
        return this.rulesLoader.isSkipHeader();
    }
    
    /** 
     * Sets skip rule header flag - <code>true</code> rule header will not be loaded, 
     * but simply set to <code>null</code>.
     * @param skipHeader value <code>true</code> means to skip rule header
     */
    public void setSkipRuleHeader(boolean skipHeader) {
        this.rulesLoader.setSkipHeader(skipHeader);        
    }
         
    /**
     * Registers user's operator under default name ({@link ClxOperatorDefFactory#getName()}).
     * @param operator factory that will create user's operator instance
     */
    public void registerOperator(ClxOperatorDefFactory operator) {
        operators.put(operator.getName(), operator);
    }
    
    /**
     * Registers user's operator under given name.
     * @param name operator name
     * @param operator factory that will create user's operator instance
     */
    public void registerOperator(String name, ClxOperatorDefFactory operator) {
        operators.put(name, operator);
    }
    
    /**
     * Post prepares all {@link ClxExtMacroCallPostPrep} calls.
     * @param macroCalls collection of {@link ClxExtMacroCall}. It may contains any implementation
     *        of macro call. Method will simply skip all calls that doesn't require post preparing.
     * @param macroMap macros binding: macro's name -&gt; macro's definition 
     *        ({@link String}-&gt;{@link ClxMacro}. 
     * @throws ClxBuilderException invalid macro call: required macro is not defined in 
     *         <code>macroMap</code>, invalid parameters in macro call
     */
    private void prepareMacros(Collection macroCalls, Map macroMap)
    throws ClxBuilderException {
        for(Iterator i = macroCalls.iterator(); i.hasNext(); ) {
            ClxExtMacroCall call = (ClxExtMacroCall) i.next();
            if (call instanceof ClxExtMacroCallPostPrep) {
                ClxExtMacroCallPostPrep toPrep = (ClxExtMacroCallPostPrep) call;
                if (! toPrep.isReady()) {
                    toPrep.postPrepare((ClxMacro) macroMap.get(toPrep.getMacroName()));
                }
            }
        }
    }
    
    /**
     * Post prepares all {@link ClxOperatorCallPostPrep} calls.
     * @param operatorCalls collection of {@link ClxOperatorCall}. It may contains any 
     *        implementation of operator call. Method will simply skip all calls that doesn't 
     *        require post preparing.
     * @param operatorMap operators binding: operator's name -&gt; operator's factory 
     *        ({@link String}-&gt;{@link ClxOperatorDefFactory}. 
     * @throws ClxBuilderException invalid operator call: required operator is not defined in 
     *         <code>operatorMap</code>, invalid parameters in the call
     */
    private void prepareOperators(Collection operatorCalls, Map operatorMap)
    throws ClxBuilderException {
        for(Iterator i = operatorCalls.iterator(); i.hasNext(); ) {
            ClxOperatorCall call = (ClxOperatorCall) i.next();
            if (call instanceof ClxOperatorCallPostPrep) {
                ClxOperatorCallPostPrep toPrep = (ClxOperatorCallPostPrep) call;
                if (! toPrep.isReady()) {
                    toPrep.postPrepare(
                            (ClxOperatorDefFactory) operatorMap.get(toPrep.getOperatorName()), 
                            this.xpathEngine);
                }
            }
        }
    }
    
    /**
     * Creates CLX validator from rule set.
     * It loads all external macros (using {@link #macrosLoader}), post prepares macro and operator
     * calls.
     * @param rules rule set
     * @return CLiXML validator; !=null
     * @throws ClxBuilderException error when loading external macros or while post preparing
     *         macro/operator calls
     */
    protected ClxValidator createValidator(ClxRules rules) 
    throws ClxBuilderException {
        //load macros
        Map macros = new HashMap();
        for(Iterator i = rules.getMacrosURL().iterator(); i.hasNext(); ) {
            String url = (String) i.next();
            ClxMacroDefinitions macroDefs;
            try {
                macroDefs = macrosLoader.loadMacroDefinitions(url);
            } catch(IOException e) {
                throw new ClxBuilderException("Error while loading macros from " + url, e); 
            }
            
            //prepare all macro calls in internal macro context
            this.prepareMacros(macroDefs.getMacroCalls(), macroDefs.getMacroMap());
            //prepare all operator calls using global operator context
            this.prepareOperators(macroDefs.getOperatorCalls(), this.operators);
            
            macros.putAll(macroDefs.getMacroMap());
        }
        
        //resolve macros
        this.prepareMacros(rules.getMacroCalls(), macros);
        this.prepareOperators(rules.getOperatorCalls(), this.operators);
        return new ClxValidator(this.xpathEngine, rules);
    }
    
    /**
     * Creates CLX validator from XML DOM element.
     * @param rulesElem XML element, this element has to have name {@link ClxConstants#TAG_RULES}
     * @return CLX validator; !=null
     * @throws ClxBuilderException error in rules definition
     */
    public ClxValidator createValidator(Element rulesElem) 
    throws ClxBuilderException {
        return this.createValidator(rulesLoader.loadRules(rulesElem));
    }
    
    /**
     * Creates CLX validator from XML DOM document.
     * @param clxDoc XML document, root element has to be {@link ClxConstants#TAG_RULES}
     * @return CLX validator; !=null
     * @throws ClxBuilderException error in rules definition
     */
    public ClxValidator createValidator(Document clxDoc) 
    throws ClxBuilderException {
        return this.createValidator(rulesLoader.loadRules(clxDoc));        
    }
    
    /**
     * Creates CLX validator from file.
     * @param file an existing file - it must be a XML file with {@link ClxConstants#TAG_RULES} 
     *        as the root
     * @return CLX validator; !=null
     * @throws ClxBuilderException error in rules definition
     * @throws IOException error while loading file
     */
    public ClxValidator createValidator(File file) 
    throws ClxBuilderException, IOException {
        return this.createValidator(rulesLoader.loadRules(file));        
    }
    
    /**
     * Creates CLX validator from stream.
     * @param is an stream ready to serve data - it must serve XML document with 
     *        {@link ClxConstants#TAG_RULES} as the root
     * @return CLX validator; !=null
     * @throws ClxBuilderException error in rules definition
     * @throws IOException error while reading data from stream
     */
    public ClxValidator createValidator(InputStream is) 
    throws ClxBuilderException, IOException {
        return this.createValidator(rulesLoader.loadRules(is));        
    }
    
    /**
     * Creates CLX validator from URL.
     * @param url URL that point to XML document with {@link ClxConstants#TAG_RULES} as the root
     * @return CLX validator; !=null
     * @throws ClxBuilderException error in rules definition
     * @throws IOException error while loading data from URL
     */
    public ClxValidator createValidator(String url) 
    throws ClxBuilderException, IOException {
        return this.createValidator(rulesLoader.loadRules(url));        
    }
            
}
