package org.xdv.clx.expr;

import org.xdv.clx.exec.ClxExecutionContext;
import org.xdv.clx.exec.ClxExecutionPoint;
import org.xdv.clx.operator.ClxValueUtil;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExpression;
import org.xdv.xpath.XPathValue;

/**
 * Binary predicate - a general CLX type.
 * It consist of exacly two xpath expressions (order is important).
 *
 * @see <a href="http://www.clixml.org/clix/1.0/docs/html/#complexType_BinaryPredicateType_Link03429858">CliXML BinaryPredicateType</a> 
 * @author Tomasz Kokoszka
 */
public abstract class ClxFormulaBinaryPredicate implements ClxFormula {
        
    /** First expression; !=null */
    private final XPathExpression op1;
    
    /** Secound expression; !=null */
    private final XPathExpression op2;
    
    /** 
     * Creates an instance.
     * @param op1 first expression
     * @param op2 secound expression
     */
    public ClxFormulaBinaryPredicate(XPathExpression op1, XPathExpression op2) {
        super();
        this.op1 = op1;
        this.op2 = op2;
    }
       
    public boolean executeFormula(ClxExecutionContext cntx, ClxExecutionPoint frame) 
    throws XdvValidationException {
        frame.finish();
        return this.executeFormula(cntx);
    }

    /**
     * Returns first expression.
     * @return first expression; !=null
     */
    public XPathExpression getOp1() {
        return op1;
    }

    /**
     * Returns secound expression.
     * @return secound expression; !=null
     */
    public XPathExpression getOp2() {
        return op2;
    }
    
    /**
     * Executes XPath expression.
     * @param xpath expression to execute
     * @param cntx CLX execution context
     * @return execution result; !=null
     * @throws XdvValidationException error while executing XPath expression
     */
    protected XPathValue executeXPath(XPathExpression xpath, ClxExecutionContext cntx) 
    throws XdvValidationException {
        try {
            return cntx.executeXPath(xpath);
        } catch(XPathException e) {
            throw new XdvValidationException("Error while executing xpath " + xpath.asString(), e);
        }        
    }
        
    /**
     * Compares results of both XPath expressions.
     * It executes both expressions and compare results (it converts result types if needed).
     * In general it returns a result of substraction: <code>op1 - op2</code>. In details:
     * <ul>
     *  <li>0 - if <code>op1</cdoe>op2</code>,
     *  <li>something less than 0 - if <code>op1</code> is smaller than <code>op2</code>,
     *  <li>something greater than 0 - if <code>op2</code> is greater than <code>op2</code>.
     * </ul>
     * @param cntx CLX execution context
     * @return comparision result 
     * @throws XdvValidationException
     */
    protected int compare(ClxExecutionContext cntx)
    throws XdvValidationException {
        return ClxValueUtil.compare(executeXPath(op1, cntx), executeXPath(op2, cntx));
    }
    
}
