package org.xdv.clx.expr;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * HEADER - CLX element implementation.
 * 
 * @see <a href="http://www.clixml.org/clix/1.0/#header">CLiXML - clx:and</a>
 * @see org.xdv.clx.base.ClxConstants#TAG_HEADER
 * @author Tomasz Kokoszka
 */
public class ClxHeader {
    
    /** 
     * List of authors - list of {@link String}; !=null.
     * It contains content of all {@link org.xdv.clx.base.ClxConstants#TAG_AUTHOR} elements from
     * this header in the same order as they are defined in XML.  
     * @see org.xdv.clx.base.ClxConstants#TAG_AUTHOR
     */
    private List authors;
    
    /** 
     * Description - list of {@link String}; !=null.
     * It contains content of all {@link org.xdv.clx.base.ClxConstants#TAG_DESCRIPTION} elements 
     * from this header in the same order as they are defined in XML.  
     * @see org.xdv.clx.base.ClxConstants#TAG_DESCRIPTION
     */
    private List descriptions;
    
    /** 
     * Comments - list of {@link String}; !=null.
     * It contains content of all {@link org.xdv.clx.base.ClxConstants#TAG_COMMENT} elements 
     * from this header in the same order as they are defined in XML.  
     * @see org.xdv.clx.base.ClxConstants#TAG_COMMENT
     */
    private List comments;
    
    /**
     * Create an empty header.
     */
    public ClxHeader() {
        super();
        this.authors = new ArrayList(0);
        this.descriptions = new ArrayList(0);
        this.comments = new ArrayList(0);
    }

    /**
     * Create a header.
     * @param authors list of authors - list of {@link String} 
     * @param descriptions descriptions - list of {@link String}
     * @param comments comments - list of {@link String}
     */
    public ClxHeader(List authors, List descriptions, List comments) {
        super();
        this.authors = authors;
        this.descriptions = descriptions;
        this.comments = comments;
    }

    /** 
     * Returns header authors - list of {@link String}.
     * @return list of {@link String}; !=null
     * @see org.xdv.clx.base.ClxConstants#TAG_AUTHOR
     */
    public List getAuthors() {
        return authors;
    }

    /** 
     * Returns header comments - list of {@link String}.
     * @return list of {@link String}; !=null
     * @see org.xdv.clx.base.ClxConstants#TAG_COMMENT
     */
    public List getComments() {
        return comments;
    }

    /** 
     * Returns header descriptions - list of {@link String}.
     * @return list of {@link String}; !=null
     * @see org.xdv.clx.base.ClxConstants#TAG_DESCRIPTION
     */
    public List getDescriptions() {
        return descriptions;
    }
    
    /**
     * Print list to a string.
     * For every element from source list:
     * <ol>
     *  <li>Take the value as a string ({@link Object#toString()}),
     *  <li>Append it to tre result string
     *  <li>If it is not the last element from list then add <code>delimer</code> to the string.
     * </ol>
     * @param list list of {@link Object}
     * @param delimer text to separate list elements in output string
     * @return printed list; !=null
     */
    private String listToString(List list, String delimer) {
        StringBuffer sb = new StringBuffer();
        for(Iterator i = list.iterator(); i.hasNext(); ) {
            sb.append(i.next());
            if (i.hasNext()) {
                sb.append(delimer);
            }
        }
        return sb.toString();
    }
    
    /** 
     * Returns header authors - all authors separated using <code>delimer</code>.
     * @param delimer text to seperate authors in result string
     * @return authors; !=null
     * @see #getAuthors()     
     */
    public String getAuthors(String delimer) {
        return listToString(authors, delimer);
    }
    
    /** 
     * Returns header descriptions - descriptions separated using <code>delimer</code>.
     * @param delimer text to seperate different description in result string
     * @return descriptions; !=null
     * @see #getDescriptions()
     */
    public String getDescriptions(String delimer) {
        return listToString(descriptions, delimer);
    }
    
    /** 
     * Returns header comments - comments separated using <code>delimer</code>.
     * @param delimer text to seperate different comments in result string
     * @return comments; !=null
     * @see #getDescriptions()
     */
    public String getComments(String delimer) {
        return listToString(comments, delimer);
    }
    
    public String toString() {
        return getDescriptions(". ");
    }
    
}
