package org.xdv.clx.extension;

import org.xdv.clx.exec.ClxExecutionContext;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExpression;

/**
 * Dynamic error message - CLX extended element implementation.
 * <br>
 * This element is dedicated to generate dynamic validation-error message.
 * Generated message is result of XPath expression converted to a string.
 *
 * @see org.xdv.clx.base.ClxConstants#TAG_EXT_VALUEOF
 * @see ClxExtErrorMessage
 * @see ClxExtErrorReport
 * @author Tomasz Kokoszka
 */
public class ClxExtErrorMessageXPath implements ClxExtErrorMessage {

    /** XPath expression - generated message is a result of this expression; !=null */
    private final XPathExpression xpath;
    
    /** Substitute static error message returned in case when <code>xpath</code>
     * cannot be executed. 
     * Value <code>null</code> means that whole evaluation should fail if expression 
     * cannot be executed.
     */
    private final String xpathErrorMsg;
    
    /**
     * Create dynamic error message.
     * @param xpath XPath expression to generate error message 
     * @param xpathErrorMsg substitute static error message returned in case when <code>xpath</code>
     *        cannot be executed
     */
    public ClxExtErrorMessageXPath(XPathExpression xpath, String xpathErrorMsg) {
        this.xpath = xpath;
        this.xpathErrorMsg = xpathErrorMsg;
    }        
    
    /**
     * Generate error message - executes XPath expression.
     * If XPath execution fails and substitute static message has not been defined then
     * method fails with {@link XdvValidationException}.
     * @param cntx CLX execution context - to execute XPath expression
     * @return error message - result of XPath expression; !=null
     * @throws XdvValidationException error while executing XPath expression.
     * @see ClxExtErrorMessage#generateMessage(org.xdv.clx.exec.ClxExecutionContext)
     */
    public String generateMessage(ClxExecutionContext cntx) 
    throws XdvValidationException {
        try {
            return cntx.executeXPath(xpath).asString();
        } catch (XPathException e) {
            if (xpathErrorMsg != null) {
                return xpathErrorMsg;
            }
            throw new XdvValidationException("XPath execution error while generating error " +
                    "message - xpath " + xpath.asString(), e);
        }
    }
    
    public String toString() {
        return "{" + xpath.asString() + "}";
    }

}
