package org.xdv.clx.extension;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.xdv.clx.exec.ClxExecutionContext;
import org.xdv.clx.exec.ClxExecutionPoint;
import org.xdv.clx.exec.ClxVariableContext;
import org.xdv.clx.expr.ClxFormula;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExpression;
import org.xdv.xpath.XPathValue;

/**
 * General MACRO CALL - CLX extended element implementation.
 * <p>
 * It is a macro call - a formula that call a predefined macro ({@link org.xdv.clx.expr.ClxMacro}).
 * This element is not defined in CLiXML 1.0. It is an extension to the standard.
 * Extension is defined in <code>clix_Macro.xsd</code>.
 * <br> 
 * Macro call is simply replaced by a call to underlying formula with given parameters. 
 * Macro defines own parameter list, macro call must define exacly the same set of parameters.
 * During the execution, macro call simply executes underlying formula in variable context that 
 * contains only macro parameters (evern rules variables are not accesible!).
 * </p>
 * 
 * @see <a href="http://www.clixml.org/clix/1.0/#macro">CLiXML - macro:*</a>
 * @see org.xdv.clx.base.ClxConstants#TAG_EXT_MACROCALL
 * @author Tomasz Kokoszka
 */
public abstract class ClxExtMacroCall implements ClxFormula {
    
    /** 
     * Macro call parameters - parameter name-&gt;expression to evaluate its value 
     * ({@link String}-&gt;{@link XPathExpression}).
     */
    protected final Map params;
    
    /**
     * Create macro call.
     * @param params call parameter - parameter name-&gt;parameter value 
     * ({@link String}-&gt;{@link XPathExpression}).
     */
    public ClxExtMacroCall(Map params) {
        this.params = params;
    }
    
    /**
     * Returns macro name.
     * @return macro name; !=null
     */
    abstract public String getMacroName();
    
    /**
     * Returns macro output - formula that will executed by this macro call.
     * <br>
     * How output is retrieved depends on implementation.  
     * @param cntx execution context
     * @return formula to execute; !=null
     * @throws XdvValidationException error when trying to get formula for this macro call
     */
    abstract protected ClxFormula getOutput(ClxExecutionContext cntx) 
    throws XdvValidationException;
    
    /**
     * Returns call parameters.
     * @return parameter name-&gt;parameter value ({@link String}-&gt;{@link XPathExpression}); 
     *         !=null
     */
    public Map getParamMap() {
        return params;
    }
    
    /**
     * Returns call parameter names.
     * @return parameter names: set of {@link String}; !=null
     */
    public Set getParams() {
        return params.keySet();
    }
    
    /**
     * Evaluates macro parameters.
     * @param cntx CLX execution context
     * @return evaluated macro parameters: parameter name-&gt;parameter value
     *         ({@link String}-&gt;{@link XPathValue}); !=null
     * @throws XdvValidationException error while evaluating parameters
     */
    private Map evaluateParams(ClxExecutionContext cntx)
    throws XdvValidationException {
        final Map result = new HashMap();
        for(Iterator i = params.entrySet().iterator(); i.hasNext(); ) {
            final Map.Entry entry = (Map.Entry) i.next();
            String name = (String) entry.getKey();
            XPathExpression xpath = (XPathExpression) entry.getValue();
            
            XPathValue value;
            try {
                value = cntx.executeXPath(xpath);
            } catch (XPathException e) {
                throw new XdvValidationException("Unable to evaluate macro parameter " + name, e);
            }
            
            result.put(name, value);            
        }
        
        return result;
    }
    
    /**
     * Adds parameters to CLX execution context.
     * @param evaluatedParams evaluated macro parameters: parameter name-&gt;parameter value
     *         ({@link String}-&gt;{@link XPathValue})
     * @param cntx CLX execution context; modified in place
     * @throws XdvValidationException error while trying to add parameter to context
     */
    private void addParamsToContext(Map evaluatedParams, ClxExecutionContext cntx) 
    throws XdvValidationException {
        for(Iterator i = evaluatedParams.entrySet().iterator(); i.hasNext(); ) {
            final Map.Entry entry = (Map.Entry) i.next();
            String name = (String) entry.getKey();
            XPathValue value = (XPathValue) entry.getValue();

            cntx.declareVariable(name);
            cntx.setVariable(name, value);
        }
    }
    
    /**
     * Removes macro parameters from CLX execution context.
     * @param cntx CLX execution context; modified in place
     * @throws XdvValidationException error while trying to remove parameter from context
     */
    private void removeParamsFromContext(ClxExecutionContext cntx) 
    throws XdvValidationException {        
        for(Iterator i = params.keySet().iterator(); i.hasNext(); ) {
            final String name = (String) i.next();
            cntx.undeclareVariable(name);
        }
    }        
    
    public boolean executeFormula(ClxExecutionContext cntx) 
    throws XdvValidationException {
        Map evaluated = evaluateParams(cntx);        
        ClxVariableContext oldVarCntx = cntx.setNewVariableContext();
        boolean b;
        try {
            addParamsToContext(evaluated, cntx);
            evaluated = null;
            b = getOutput(cntx).executeFormula(cntx);
            removeParamsFromContext(cntx);
        } finally {
            cntx.setNewVariableContext(oldVarCntx);
        }
        return b;
    }
    
    public boolean executeFormula(ClxExecutionContext cntx, ClxExecutionPoint frame) 
    throws XdvValidationException {
        Map evaluated = evaluateParams(cntx);        
        ClxVariableContext oldVarCntx = cntx.setNewVariableContext();
        boolean b;
        try {
            addParamsToContext(evaluated, cntx);
            evaluated = null;
            b = getOutput(cntx).executeFormula(cntx, frame.getNext1());        
            removeParamsFromContext(cntx);
        } finally {
            cntx.setNewVariableContext(oldVarCntx);
        }
        
        frame.setFinished();
        return b;
    }
    
    /**
     * Returns string containing parameter names separated by comma sourounded by brackets.
     * For instance: <code>(base, factor)</code>. 
     * @return string containing parameter names; !=null
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("(");
        for(Iterator i = params.keySet().iterator(); i.hasNext(); ) {
            sb.append(i.next());
            if (i.hasNext()) {
                sb.append(", ");
            }
        }
        sb.append(")");
        
        return sb.toString();
    }
    
}
