package org.xdv.clx.operator;

import java.util.List;

import org.xdv.clx.builder.ClxBuilderException;
import org.xdv.clx.exec.ClxExecutionContext;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathEngine;
import org.xdv.xpath.XPathException;

/**
 * Post prepared OPERATOR CALL - CLX element implementation.
 * <p>
 * This call can be created before definition of called operator known. But before first execution
 * one must <i>post prepare</i> this call by simply specifying required operator definition.
 * </p>
 * 
 * @author Tomasz Kokoszka
 */
public class ClxOperatorCallPostPrep extends ClxOperatorCall {
   
    /** Operator name */
    protected final String name;
    
    /** Call parameters - list of {@link ClxOperatorParam}; !=null */
    protected final List params;
    
    /**
     * Operator definition that will be called to evaluate call value.
     * Value <code>null</code> means that this call has not been prepared to be executed yet.
     */
    protected ClxOperatorDef operatorDef = null;
    
    /**
     * Create call - this call is not ready to execute yet, it requires <i>post preparing</i>.
     * @param name operator name
     * @param params list of call parameters - list of {@link ClxOperatorParam}
     */
    public ClxOperatorCallPostPrep(String name, List params) {
        this.name = name;
        this.params = params;
    }
    
    public String getOperatorName() {
        return this.name;
    }
    
    /**
     * Is call ready to execution?
     * Call is ready if has been <i>post prepared</i> 
     * (see {@link #postPrepare(ClxOperatorDefFactory, XPathEngine)}).
     * @return value <code>true</code> if this call has been post prepared and is ready to execute
     */
    public boolean isReady() {
        return this.operatorDef != null;
    }
    
    /**
     * Post prepares this operator call.
     * It instantiates user operator using <code>factory</code>. 
     * <br>
     * After calling this method, {@link #isReady()} will return <code>true</code>.
     * @param factory factory to instantiate operator definition
     * @param engine XPath engine for operator factory to evaluate parameters (is needed)
     * @return previous operator instance used by this call, 
     *         <code>null</code> means that this call has not been ready for executing before  
     * @throws ClxBuilderException factory is empty (<code>null</code>) or factory cannot create
     *         instance using this call parameters
     */
    public ClxOperatorDef postPrepare(ClxOperatorDefFactory factory, XPathEngine engine) 
    throws ClxBuilderException {        
        if (factory == null) {
            throw new ClxBuilderException("Operator " + getOperatorName() + 
                    " called but not defined");
        }        
        assert factory != null;                       
        
        ClxOperatorDef old = this.operatorDef;
        try {
            this.operatorDef = factory.createCall(this.params, engine);
        } catch (XPathException e) {
            throw new ClxBuilderException("Error when compiling xpath", e);
        }
        return old;
    }
           
    public boolean executeFormula(ClxExecutionContext cntx) 
    throws XdvValidationException {
        if (operatorDef == null) {
            throw new XdvValidationException("Operator call " + name + " has not been " +
                    "prepared for execution");
        }

        return operatorDef.executeFormula(cntx);
    }
    
    public String toString() {
        return "operatorCall:postPrepare(" + name + ")(" + operatorDef + ")";
    }

}
