package org.xdv.clx.operator;

import org.xdv.clx.exec.ClxExecutionContext;
import org.xdv.clx.exec.ClxExecutionPoint;
import org.xdv.clx.expr.ClxFormula;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExpression;
import org.xdv.xpath.XPathValue;

/**
 * Plugin operator deifnition - sekeleton to implement operator.
 * <p>
 * Plugin operators are scripted predicates that can take a number of parameters and 
 * return <code>true</code> or <code>false</code>. They can be used to perform a variety of tasks 
 * like database lookups, communication with legacy systems or extensive computation.
 * </p>
 * <p>
 * To define a new plugin operator:
 * <ol>
 *  <li>Implement operator behavior by extending this class and implementing method
 *      {@link #executeFormula(ClxExecutionContext)}.
 *  <li>Create a factory to instantiate plugin operator - {@link ClxOperatorDefFactory}.
 *  <li>Register the factory in validator factory - 
 *      {@link org.xdv.clx.base.ClxValidatorFactory#registerOperator(ClxOperatorDefFactory)}.
 * </ol>
 * Class {@link ClxXPathCondition} is a simple implementation of plugin operator (definition
 * in a private class). 
 * You should take a look at it before creating your first operator.
 * </p>
 * 
 * @see <a href="http://www.clixml.org/clix/1.0/#operator">Extension mechanisms</a>
 * @author Tomasz Kokoszka
 */
public abstract class ClxOperatorDef implements ClxFormula {
    
    /** Creates operator */
    public ClxOperatorDef() {
    }
    
    /**
     * Execute/evaluate operator value.
     * <br>Executes formula that stays behind this operator.
     * @param cntx CLX execution context - variable binding, convinient methods to execute XPath
     *        expressions
     * @return evaluated value
     * @throws XdvValidationException any error during evaluation - invalid XPath, invalid state, 
     *         etc.
     */
    abstract public boolean executeFormula(ClxExecutionContext cntx)
    throws XdvValidationException;
    
    public final boolean executeFormula(ClxExecutionContext cntx, ClxExecutionPoint frame) 
    throws XdvValidationException {
        frame.finish();
        return executeFormula(cntx);
    }
    
    /**
     * Convenient method to execute XPath expression.
     * It is a link to {@link ClxExecutionContext#executeXPath(XPathExpression)}.
     * @param xpath XPath expression to execute
     * @param cntx CLX execution context
     * @return evaluation result; !=null
     * @throws XdvValidationException error while executin xpath
     */
    protected XPathValue executeXPath(XPathExpression xpath, ClxExecutionContext cntx) 
    throws XdvValidationException {
        try {
            return cntx.executeXPath(xpath);
        } catch(XPathException e) {
            throw new XdvValidationException("Error while executing xpath " + xpath.asString() + 
                    " for operator", e);
        }
    }
    
}
