package org.xdv.clx.operator;

import org.xdv.xpath.XPathValue;

/**
 * Set of static methods to operate on XPath value in context of executing CLiXML expressions.
 * <br>
 * Provides functionality to compare and convert values according to CLiXML conversion rules 
 * defined in CLiXML specification:
 * <ul>
 *  <li><a href="http://www.clixml.org/clix/1.0/#nodeset-conversion">Node set conversion</a>
 *      - rules for nodes and node sets,
 *  <li><a href="http://www.clixml.org/clix/1.0/#equal">clix:equal</a> - rules for converting
 *      types when comparing their values.
 * </ul>
 *
 * @author Tomasz Kokoszka
 */
public class ClxValueUtil {
    
    /**
     * Converts number to boolean.
     * <br>
     * A non-zero number except <code>NaN</code> is converted to <code>true</code>. 
     * <code>0</code> and <code>NaN</code> are converted to <code>false</code>.
     * @param v number value (have to be of type {@link XPathValue#TYPE_NUMBER})
     * @return result of conversion
     */
    static public boolean number2boolean(XPathValue v) {
        assert v.getType() == XPathValue.TYPE_NUMBER;
        double d = v.asNumber();
        return (! Double.isNaN(d)) && (d != 0);
    }
    
    /**
     * Converts number to string.
     * <br>
     * Number is converted to a a text using IEEE754 rule:
     * <code>NaN</code> is converted to 'NaN', <code>0.0</code> to '0', <code>5.23400</code> 
     * to '5.234' etc.
     * @param v text value (have to be of type {@link XPathValue#TYPE_STRING})
     * @return result of conversion; !=null
     */
    static public String number2string(XPathValue v) {
        assert v.getType() == XPathValue.TYPE_NUMBER;
        return Double.toString(v.asNumber());
    }
        
    /**
     * Converts boolean to string.
     * <br>
     * Value <code>true</code> is converted to 'true', <code>false</code> to 'false'.
     * @param v boolean value (have to be of type {@link XPathValue#TYPE_BOOLEAN})
     * @return result fo conversion; !=null
     */
    static public String boolean2string(XPathValue v) {
        assert v.getType() == XPathValue.TYPE_BOOLEAN;
        return Boolean.toString(v.asBoolean());
    }
    
    /**
     * Converts node set to a string.
     * <br>
     * Conversion according to CLiXML specification: 
     * <a href="http://www.clixml.org/clix/1.0/#nodeset-conversion">Node set conversion</a>.
     * @param v node set value (have to be of type {@link XPathValue#TYPE_NODESET})
     * @return result of conversion; !=null
     */
    static public String nodeset2string(XPathValue v) {
        return v.asString();
    }
    
    /**
     * Compares two booleans.
     * @param v1 first boolean value
     * @param v2 secound boolean value
     * @return result of <code>v1-v2</code>: <ul>
     *          <li><code>0</code> = <code>v1=v2</code>
     *          <li><code>-1</code> = <code>v1&lt;v2</code>
     *          <li><code>1</code> = <code>v1&gt;v2</code>
     *          </ul> 
     */
    static private int compareBoolean(boolean v1, boolean v2) {
        return (v1?1:0) - (v2?1:0);
    }
    
    /**
     * Compares two numbers.
     * @param v1 first number 
     * @param v2 secound number
     * @return result of <code>v1-v2</code>
     */
    static private int compareNumber(double v1, double v2) {
        return Double.compare(v1, v2);
    }
    
    /**
     * Compares two strings lexicographically.
     * @param v1 first string 
     * @param v2 secound string
     * @return result of <code>v1.compareTo(v2)</code> - see {@link String#compareTo(String)} 
     */
    static private int compareString(String v1, String v2) {
        return v1.compareTo(v2);
    }
    
    /**
     * Compares two values accordign to CLiXML comparistion and conversion rules. 
     * @param v1 first value
     * @param v2 secound value
     * @return result of <code>v1-v2</code>. Value <code>0</code> if both arguments are equal; 
     * a value less than <code>0</code> if <code>v1&lt;v2</code>; 
     * and a value greater than <code>0</code> if <code>v1&gt;v2</code>.
     * @see <a href="http://www.clixml.org/clix/1.0/#nodeset-conversion">CLiXML - node set conversion</a>
     * @see <a href="http://www.clixml.org/clix/1.0/#equal">clix:equal - comparing values</a>
     */
    static public int compare(XPathValue v1, XPathValue v2) {        
        switch (v1.getType()) {

        case XPathValue.TYPE_BOOLEAN:
            switch (v2.getType()) {
            case XPathValue.TYPE_BOOLEAN:
                return compareBoolean(
                        v1.asBoolean(),
                        v2.asBoolean());

            case XPathValue.TYPE_NUMBER:
                return compareBoolean(
                        v1.asBoolean(),
                        number2boolean(v2));
                
            case XPathValue.TYPE_NODESET:
                return compareString(
                        boolean2string(v1),
                        nodeset2string(v2));
                
            default:
                return compareString(
                        boolean2string(v1),
                        v2.asString());
            }
            
        case XPathValue.TYPE_NUMBER:
            switch (v2.getType()) {
            case XPathValue.TYPE_BOOLEAN:
                return compareBoolean(
                        number2boolean(v2),
                        v2.asBoolean());
                
            case XPathValue.TYPE_NUMBER:
                return compareNumber(
                        v1.asNumber(),
                        v2.asNumber());
                
            case XPathValue.TYPE_NODESET:
                return compareString(
                        number2string(v1),
                        nodeset2string(v2));

            default:
                return compareString(
                        number2string(v1),
                        v2.asString());
            }                  
            
        default:
            String str1;
            if (v1.getType() == XPathValue.TYPE_NODESET) {
                str1 = nodeset2string(v1);
            } else {
                str1 = v1.asString();
            }
            switch (v2.getType()) {
            case XPathValue.TYPE_BOOLEAN:
                return compareString(str1, boolean2string(v2));
                
            case XPathValue.TYPE_NUMBER:
                return compareString(str1, number2string(v2));
                
            case XPathValue.TYPE_NODESET:
                return compareString(str1, nodeset2string(v2));
                
            default:
                return compareString(str1, v2.asString());
            }
        }
    }


}
