package org.xdv.xpath;

import java.util.Collection;

/**
 * Entry point to use XPath implementation, acts like a factory for all other components.
 * <p>
 * From user point of view it is an entry point to use specific XPath implementation.
 * It acts as a factory to:
 * <ul class="NoTopMargin">
 * <li>create xpath expression - {@link #createExpression(String)},
 * <li>create execution unit - {@link #createExecutionUnit(Object)},
 * <li>create variable map - {@link #createVariableMap()},
 * <li>create xpath value - {@link #createValue(Object)}, 
 *     {@link #createValueBoolean(boolean)} etc.
 * </ul>    
 * User should always use factory instead of creating objects manually.
 * </p>
 * <p>
 * Every XPath implementation expects XML input data in specific format 
 * (W3C DOM, JDOM, etc.). 
 * For instance if underlying XPath implementation works with W3C DOM source, then it will 
 * require a {@link org.w3c.dom.Document} instance as XML document and {@link org.w3c.dom.Node} 
 * instance as XML node.
 * Wrapper takes those data as a general objects ({@link java.lang.Object}), and casts to required class. 
 * If user pass data in different format then {@link org.xdv.xpath.XPathInvalidTypeError}
 * is thrown.
 * <br> 
 * User may check accepted type of XML document by calling {@link #getDocumentClass()}.
 * </p>  
 *
 * @author Tomasz Kokoszka
 */
public interface XPathEngine {
    
    /**
     * Returns type of accepted XML document - class of XML documents that this XPath 
     * implementation works with. 
     * @return Type of XML document acceptable by this implementation; !=null 
     */
    public Class getDocumentClass();
    
    /**
     * Returns type of accepted XML node - class of XML node that this XPath 
     * implementation works with. 
     * @return Type of XML node acceptable by this implementation; !=null 
     */
    public Class getNodeClass();
    
    /**
     * Returns node text value.
     * Value depends on node type:
     * <ol class="NoTopMargin">
     * <li>element - concatenated value of all text nodes children of the element
     * <li>attribute - value of attribute
     * <li>text node - content of text node
     * <li>comment - content of the comment, excluding the <code>&lt;!--</code> and
     *     <code>--&gt;</code> escape characters
     * <li>processing instruction - content of the processing instruction following the target,
     *     but excluding the terminating <code>?&gt;</code>
     * </ol>
     * @param nodeNative XML node
     * @return Node text value; !=null
     * @throws XPathInvalidTypeError invalid XML node class, an instance of 
     *         {@link #getNodeClass()} was expected
     */
    public String getNodeString(Object nodeNative)
    throws XPathInvalidTypeError;

    /**
     * Compares two nodes - returns true if those are exactly the same nodes.
     * @param nodeNative1 XML node
     * @param nodeNative2 XML node
     * @return Value <code>true</code> if and only if both objects represents exatly the same 
     *         XML node (from the same XML document)
     * @throws XPathInvalidTypeError invalid XML node class, an instance of 
     *         {@link #getNodeClass()} was expected
     */
    public boolean isSameNode(Object nodeNative1, Object nodeNative2)
    throws XPathInvalidTypeError;

    /**
     * Creates XPath expression.
     * It may be a precompiled expression or it a simple text wrapper - 
     * it depends on implementation.
     * @param xpath XPath expression as text, for instance: 
     *        <code>/data/field[substring(@name, 1, 2)='In']</code>
     * @return XPath expression; !=null 
     * @throws XPathException error while trying to create expression 
     *         (xpath expression is not valid), 
     */
    public XPathExpression createExpression(String xpath) 
    throws XPathException;
    
    /**
     * Creates XPath execution unit.
     * Unit is connected to a XML document. This binding is final and cannot be changed.
     * @param documentNative source XML document. 
     *        Exact type depends on implementation, if type is invalid then 
     *        a {@link XPathInvalidTypeError} is thrown. 
     * @return Execution unit; !=null
     * @throws XPathInvalidTypeError invalid XML document class, an instance of 
     *         {@link #getDocumentClass()} was expected
     */
    public XPathExecutionUnit createExecutionUnit(Object documentNative)
    throws XPathInvalidTypeError;
    
    /**
     * Creates an empty variable map.
     * @return New, empty variable map; !=null. Some implementations may put
     *         some global bindings there.
     */
    public XPathVariableMap createVariableMap();
    
    /**
     * Creates wrapper for XPath value.
     * @param valueNative native value, its type is specific for implementation.
     *        For instance it could be {@link String}, {@link Integer}, {@link org.w3c.dom.NodeList}
     *        - it depends of underlying XPath implementation.        
     * @return Wrapper for given native value; !=null
     * @throws XPathException unable to create wrapper, type of native value is not supported 
     *         or value is simply not valid
     */
    public XPathValue createValue(Object valueNative) 
    throws XPathException;
    
    /**
     * Creates wrapper for <b>text</b> value.
     * @param value a text 
     * @return Wrapper that represents given text; !=null
     */
    public XPathValue createValueString(String value);
        
    /**
     * Creates wrapper for <b>number</b>.
     * @param value a number 
     * @return Wrapper that represents given number; !=null
     */
    public XPathValue createValueNumber(double value);
    
    /**
     * Creates wrapper for <b>boolean</b> value.
     * @param value a boolean
     * @return Wrapper that represents given boolean value; !=null
     */    
    public XPathValue createValueBoolean(boolean value);
    
    /**
     * Creates wrapper for <b>single XML node</b>.
     * <br>
     * Implementation expects native data in correct format.
     * If wrong type of XML node is passed then a {@link XPathInvalidTypeError} will be thrown.  
     * @param nodeNative a XML node
     * @return Wrapper that represents given node - set constaining exactly one node; !=null
     * @throws XPathInvalidTypeError invalid XML node type
     */
    public XPathValueNodeSet createValueNode(Object nodeNative)
    throws XPathInvalidTypeError;
    
    /**
     * Creates wrapper for <b>XML node-set</b>.
     * <br>
     * Implementation expects native data in correct format.
     * If wrong type of XML node is passed then a {@link XPathInvalidTypeError} will be thrown.  
     * @param nodesNative collection of native nodes ({@link Object}) 
     * @return Wrapper that represents given node - set constaining exactly one node; !=null
     * @throws XPathInvalidTypeError invalid XML node type
     */
    public XPathValueNodeSet createValueNodeSet(Collection nodesNative)
    throws XPathInvalidTypeError;
    
    
    
}
