package org.xdv.xpath;

import java.util.List;

/**
 * XPath value - node set. 
 * <p>
 * Instance should never be created directly but through {@link XPathEngine}, or as a result 
 * of XPath expression execution.
 * </p>
 *
 * @see XPathValue
 * @author Tomasz Kokoszka
 */
public interface XPathValueNodeSet extends XPathValue {
       
    /**
     * Returns list of XML nodes from this node-set.
     * <br>
     * List should keep the order. It means that that nodes in list should be in the same
     * order as they were put in by XPath expression. 
     * <br>
     * List should not be modified after retrieved - some implementations may
     * relax this constraint.
     * @return List of XML nodes; !=null. If node set is empty, then an empty list is returned.
     *         Node class depends on implementation. This list should not modified 
     *         (on write access allowed). 
     */
    public List getNodes();

    /**
     * Returns list of wrapped XML nodes -
     * list of {@link XPathValueNodeSet}, each list element is a node-set
     * containing single XML node from this node-set.
     * <br>
     * List should keep the order. It means that that nodes in list should be in the same
     * order as they were put in by XPath expression. 
     * <br>
     * List should not be modified after retrieved - some implementations may
     * relax this constraint. 
     * @return List of {@link XPathValueNodeSet}; !=null. 
     *         If node set is empty, then an empty list is returned. 
     */
    public List getNodesAsValues();
        
    /**
     * Returns the first XML node from this node-set.
     * @return First XML node (in document order) or <code>null</code> if this node-set is empty.
     *         Node class depends on implementation.
     */
    public Object getFirstNode();
    
    /**
     * Returns <code>true</code> if this node-set contains the specified XML node. 
     * @param nodeNative XML node whose presence in this node-set is to be tested
     * @return Value <code>true</code> if and only if this node-set contains the specified XML node
     * @throws XPathInvalidTypeError invalid XML node class (class depends on implementation)
     */
    public boolean containsNode(Object nodeNative)
    throws XPathInvalidTypeError;
    
    /**
     * Returns number of XML nodes in this node-set.
     * @return Number of nodes in this node-set. 0=empty
     */
    public int getLength();
    
    /**
     * Returns <code>true</code> if and only if node-set is empty.
     * @return Value <code>true</code> if and ony if node-set is empty.
     */
    public boolean isEmpty();


}
